from unittest import mock
import unittest

import django
from django.contrib.admin.models import LogEntry, CHANGE, DELETION
from django.contrib.admin.sites import AdminSite
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.test.utils import override_settings
from waffle import get_waffle_flag_model
from waffle.admin import (FlagAdmin, SwitchAdmin, InformativeManyToManyRawIdWidget, enable_for_all,
                          disable_for_all, delete_individually, enable_switches, disable_switches)
from waffle.models import Switch
from waffle.tests.base import TestCase
from waffle.utils import get_setting


django_version = tuple(int(d) for d in django.get_version().split("."))


class FakeSuperUser:
    def has_perm(self, perm):
        return True


class FakeRequest:
    def __init__(self):
        self.GET = {}
        self.user = get_user_model().objects.create(username="test1")


Flag = get_waffle_flag_model()


skip_if_admin_permissions_not_available = \
    unittest.skipIf(django_version < (2, 1, 0), "Feature not available")


class FlagAdminTests(TestCase):
    def setUp(self):
        super().setUp()
        self.site = AdminSite()
        self.flag_admin = FlagAdmin(Flag, self.site)

    def test_informative_widget(self):
        request = mock.Mock()
        request.has_perm = lambda self, perm: True
        form = self.flag_admin.get_form(request)()
        user_widget = form.fields["users"].widget

        self.assertIsInstance(user_widget, InformativeManyToManyRawIdWidget)
        user1 = get_user_model().objects.create(username="test1")
        user2 = get_user_model().objects.create(username="test2")
        self.assertIn("(test1, test2)",
                      user_widget.render("users", [user1.pk, user2.pk]))

    def test_enable_for_all(self):
        f1 = Flag.objects.create(name="flag1", everyone=False)

        request = FakeRequest()
        enable_for_all(None, request, Flag.objects.all())

        f1.refresh_from_db()
        self.assertTrue(f1.everyone)
        log_entry = LogEntry.objects.get(user=request.user)
        self.assertEqual(log_entry.action_flag, CHANGE)
        self.assertEqual(log_entry.object_repr, "flag1 on")

    def test_disable_for_all(self):
        f1 = Flag.objects.create(name="flag1", everyone=True)

        request = FakeRequest()
        disable_for_all(None, request, Flag.objects.all())

        f1.refresh_from_db()
        self.assertFalse(f1.everyone)
        log_entry = LogEntry.objects.get(user=request.user)
        self.assertEqual(log_entry.action_flag, CHANGE)
        self.assertEqual(log_entry.object_repr, "flag1 off")

    def test_delete_individually(self):
        Flag.objects.create(name="flag1", everyone=True)

        request = FakeRequest()
        delete_individually(None, request, Flag.objects.all())

        self.assertIsNone(Flag.objects.first())
        log_entry = LogEntry.objects.get(user=request.user)
        self.assertEqual(log_entry.action_flag, DELETION)

    @skip_if_admin_permissions_not_available
    def test_flag_no_actions_without_permissions(self):
        request = FakeRequest()
        actions = self.flag_admin.get_actions(request)

        self.assertEqual(actions.keys(), set())

    @skip_if_admin_permissions_not_available
    def test_flag_action_change(self):
        request = FakeRequest()
        request.user.user_permissions.add(Permission.objects.get(codename="change_flag"))
        actions = self.flag_admin.get_actions(request)

        self.assertEqual(actions.keys(), {"enable_for_all", "disable_for_all"})

    @skip_if_admin_permissions_not_available
    def test_flag_action_delete(self):
        request = FakeRequest()
        request.user.user_permissions.add(Permission.objects.get(codename="delete_flag"))
        actions = self.flag_admin.get_actions(request)

        self.assertEqual(actions.keys(), {"delete_individually"})

    def test_model_can_be_registered_by_default(self):
        config = get_setting("ENABLE_ADMIN_PAGES")
        _register_model_to_admin_site(admin_site=self.site, config_setting=config, model=Flag)
        self.assertTrue(self.site.is_registered(Flag))

    @override_settings(WAFFLE_ENABLE_ADMIN_PAGES=False)
    def test_admin_page_can_be_disabled(self):
        config = get_setting("ENABLE_ADMIN_PAGES")
        _register_model_to_admin_site(admin_site=self.site, config_setting=config, model=Flag)
        self.assertFalse(self.site.is_registered(Flag))


class SwitchAdminTests(TestCase):
    def setUp(self):
        super().setUp()
        self.site = AdminSite()
        self.switch_admin = SwitchAdmin(Switch, self.site)

    def test_enable_switches(self):
        s1 = Switch.objects.create(name="switch1", active=False)

        request = FakeRequest()
        enable_switches(None, request, Switch.objects.all())

        s1.refresh_from_db()
        self.assertTrue(s1.active)
        log_entry = LogEntry.objects.get(user=request.user)
        self.assertEqual(log_entry.action_flag, CHANGE)
        self.assertEqual(log_entry.object_repr, "switch1 on")

    def test_disable_switches(self):
        s1 = Switch.objects.create(name="switch1", active=True)

        request = FakeRequest()
        disable_switches(None, request, Switch.objects.all())

        s1.refresh_from_db()
        self.assertFalse(s1.active)
        log_entry = LogEntry.objects.get(user=request.user)
        self.assertEqual(log_entry.action_flag, CHANGE)
        self.assertEqual(log_entry.object_repr, "switch1 off")

    @skip_if_admin_permissions_not_available
    def test_switch_no_actions_without_permissions(self):
        request = FakeRequest()
        actions = self.switch_admin.get_actions(request)

        self.assertEqual(actions.keys(), set())

    @skip_if_admin_permissions_not_available
    def test_switch_action_change(self):
        request = FakeRequest()
        request.user.user_permissions.add(Permission.objects.get(codename="change_switch"))
        actions = self.switch_admin.get_actions(request)

        self.assertEqual(actions.keys(), {"enable_switches", "disable_switches"})

    @skip_if_admin_permissions_not_available
    def test_switch_action_delete(self):
        request = FakeRequest()
        request.user.user_permissions.add(Permission.objects.get(codename="delete_switch"))
        actions = self.switch_admin.get_actions(request)

        self.assertEqual(actions.keys(), {"delete_individually"})

    def test_model_can_be_registered_by_default(self):
        config = get_setting("ENABLE_ADMIN_PAGES")
        _register_model_to_admin_site(admin_site=self.site, config_setting=config, model=Switch)
        self.assertTrue(self.site.is_registered(Switch))

    @override_settings(WAFFLE_ENABLE_ADMIN_PAGES=False)
    def test_admin_page_can_be_disabled(self):
        config = get_setting("ENABLE_ADMIN_PAGES")
        _register_model_to_admin_site(admin_site=self.site, config_setting=config, model=Switch)
        self.assertFalse(self.site.is_registered(Switch))


def _register_model_to_admin_site(admin_site, config_setting, model):
    if config_setting:
        admin_site.register(model)



from typing import TYPE_CHECKING, Any, Generic, TypeVar

from django.db import models

from waffle.utils import get_setting, get_cache


if TYPE_CHECKING:
    from waffle.models import _BaseModelType, AbstractBaseFlag, AbstractBaseSample, AbstractBaseSwitch  # noqa: F401
else:
    _BaseModelType = TypeVar("_BaseModelType")


class BaseManager(models.Manager, Generic[_BaseModelType]):
    KEY_SETTING = ''

    def get_by_natural_key(self, name: str) -> _BaseModelType:
        return self.get(name=name)

    def create(self, *args: Any, **kwargs: Any) -> _BaseModelType:
        cache = get_cache()
        ret = super().create(*args, **kwargs)
        cache_key = get_setting(self.KEY_SETTING)
        cache.delete(cache_key)
        return ret


class FlagManager(BaseManager['AbstractBaseFlag']):
    KEY_SETTING = 'ALL_FLAGS_CACHE_KEY'


class SwitchManager(BaseManager['AbstractBaseSwitch']):
    KEY_SETTING = 'ALL_SWITCHES_CACHE_KEY'


class SampleManager(BaseManager['AbstractBaseSample']):
    KEY_SETTING = 'ALL_SAMPLES_CACHE_KEY'



from typing import Any

from argparse import ArgumentTypeError
from django.core.management.base import BaseCommand, CommandError, CommandParser

from waffle import get_waffle_switch_model


def on_off_bool(string: str) -> bool:
    if string not in ['on', 'off']:
        raise ArgumentTypeError(f"invalid choice: {string!r} (choose from 'on', "
                                "'off')")
    return string == 'on'


class Command(BaseCommand):
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            'name',
            nargs='?',
            help='The name of the switch.')
        parser.add_argument(
            'state',
            nargs='?',
            type=on_off_bool,
            help='The state of the switch: on or off.')
        parser.add_argument(
            '-l', '--list',
            action='store_true',
            dest='list_switches',
            default=False,
            help='List existing switches.')
        parser.add_argument(
            '--create',
            action='store_true',
            dest='create',
            default=False,
            help='If the switch does not exist, create it.')

    help = 'Activate or deactivate a switch.'

    def handle(self, *args: Any, **options: Any) -> None:
        if options['list_switches']:
            self.stdout.write('Switches:')
            for switch in get_waffle_switch_model().objects.iterator():
                self.stdout.write(
                    '{}: {}'.format(switch.name, 'on' if switch.active else 'off')
                )
            self.stdout.write('')
            return

        switch_name = options['name']
        state = options['state']

        if not (switch_name and state is not None):
            raise CommandError('You need to specify a switch name and state.')

        if options["create"]:
            switch, created = get_waffle_switch_model().objects.get_or_create(
                name=switch_name
            )
            if created:
                self.stdout.write(f'Creating switch: {switch_name}')
        else:
            try:
                switch = get_waffle_switch_model().objects.get(name=switch_name)
            except get_waffle_switch_model().DoesNotExist:
                raise CommandError('This switch does not exist.')

        switch.active = state
        switch.save()



from django.contrib.auth.models import AnonymousUser
from django.http import Http404
from django.test import RequestFactory

from test_app import views
from waffle.middleware import WaffleMiddleware
from waffle.models import Flag, Sample, Switch
from waffle.tests.base import TestCase


def get(**kw):
    request = RequestFactory().get('/foo', data=kw)
    request.user = AnonymousUser()
    return request


def get_response(request):
    return 'hello'


def process_request(request, view):
    response = view.as_view()(request)
    return WaffleMiddleware(get_response).process_response(request, response)


class WaffleFlagMixinTest(TestCase):
    def setUp(self):
        super().setUp()
        self.request = get()

    def test_flag_must_be_active(self):
        view = views.FlagView
        self.assertRaises(Http404, process_request, self.request, view)

        Flag.objects.create(name='foo', everyone=True)
        response = process_request(self.request, view)
        self.assertEqual(b'foo', response.content)

    def test_flag_must_be_inactive(self):
        view = views.FlagOffView
        response = process_request(self.request, view)
        self.assertEqual(b'foo', response.content)

        Flag.objects.create(name='foo', everyone=True)
        self.assertRaises(Http404, process_request, self.request, view)

    def test_override_with_cookie(self):
        Flag.objects.create(name='foo', percent='0.1')
        self.request.COOKIES['dwf_foo'] = 'True'
        response = process_request(self.request, views.FlagView)
        self.assertEqual(b'foo', response.content)
        self.assertIn('dwf_foo', response.cookies)
        self.assertEqual('True', response.cookies['dwf_foo'].value)


class WaffleSampleMixinTest(TestCase):
    def setUp(self):
        super().setUp()
        self.request = get()

    def test_sample_must_be_active(self):
        view = views.SampleView
        self.assertRaises(Http404, process_request, self.request, view)

        Sample.objects.create(name='foo', percent='100.0')
        response = process_request(self.request, view)
        self.assertEqual(b'foo', response.content)

    def test_sample_must_be_inactive(self):
        view = views.SampleOffView
        response = process_request(self.request, view)
        self.assertEqual(b'foo', response.content)

        Sample.objects.create(name='foo', percent='100.0')
        self.assertRaises(Http404, process_request, self.request, view)

    def test_override_with_cookie(self):
        Sample.objects.create(name='foo', percent='0.0')
        self.request.COOKIES['dwf_foo'] = 'True'
        self.assertRaises(Http404, process_request, self.request,
                          views.SwitchView)


class WaffleSwitchMixinTest(TestCase):
    def setUp(self):
        super().setUp()
        self.request = get()

    def test_switch_must_be_active(self):
        view = views.SwitchView
        self.assertRaises(Http404, process_request, self.request, view)

        Switch.objects.create(name='foo', active=True)
        response = process_request(self.request, view)
        self.assertEqual(b'foo', response.content)

    def test_switch_must_be_inactive(self):
        view = views.SwitchOffView
        response = process_request(self.request, view)
        self.assertEqual(b'foo', response.content)

        Switch.objects.create(name='foo', active=True)
        self.assertRaises(Http404, process_request, self.request, view)

    def test_no_override_with_cookie(self):
        Switch.objects.create(name='foo', active=False)
        self.request.COOKIES['dwf_foo'] = 'True'
        self.assertRaises(Http404, process_request, self.request,
                          views.SwitchView)



from django.test import TestCase

from waffle import (
    get_waffle_flag_model,
    get_waffle_sample_model,
    get_waffle_switch_model,
)
from django.contrib.auth.models import User


class ModelsTests(TestCase):
    def test_natural_keys(self):
        flag = get_waffle_flag_model().objects.create(name='test-flag')
        switch = get_waffle_switch_model().objects.create(name='test-switch')
        sample = get_waffle_sample_model().objects.create(name='test-sample', percent=0)

        self.assertEqual(flag.natural_key(), ('test-flag',))
        self.assertEqual(switch.natural_key(), ('test-switch',))
        self.assertEqual(sample.natural_key(), ('test-sample',))

        self.assertEqual(
            get_waffle_flag_model().objects.get_by_natural_key("test-flag"), flag
        )
        self.assertEqual(
            get_waffle_switch_model().objects.get_by_natural_key("test-switch"), switch
        )
        self.assertEqual(
            get_waffle_sample_model().objects.get_by_natural_key("test-sample"), sample
        )

    def test_flag_is_not_active_for_none_requests(self):
        flag = get_waffle_flag_model().objects.create(name='test-flag')
        self.assertEqual(flag.is_active(None), False)

    def test_is_active_for_user_when_everyone_is_active(self):
        flag = get_waffle_flag_model().objects.create(name='test-flag')
        flag.everyone = True
        self.assertEqual(flag.is_active_for_user(User()), True)



from __future__ import annotations

from functools import partial

from django.http import Http404

from waffle import switch_is_active, flag_is_active, sample_is_active


class BaseWaffleMixin:

    def validate_waffle(self, waffle, func):
        if waffle.startswith('!'):
            active = not func(waffle[1:])
        else:
            active = func(waffle)
        return active

    def invalid_waffle(self):
        raise Http404('Inactive waffle')


class WaffleFlagMixin(BaseWaffleMixin):
    """
    Checks that as flag is active, or 404. Operates like the FBV decorator
    waffle_flag
    """

    waffle_flag: str | None = None

    def dispatch(self, request, *args, **kwargs):
        func = partial(flag_is_active, request)
        active = self.validate_waffle(self.waffle_flag, func)

        if not active:
            return self.invalid_waffle()

        return super().dispatch(request, *args, **kwargs)


class WaffleSampleMixin(BaseWaffleMixin):
    """
    Checks that as switch is active, or 404. Operates like the FBV decorator
    waffle_sample.
    """

    waffle_sample: str | None = None

    def dispatch(self, request, *args, **kwargs):
        active = self.validate_waffle(self.waffle_sample, sample_is_active)

        if not active:
            return self.invalid_waffle()

        return super().dispatch(request, *args, **kwargs)


class WaffleSwitchMixin(BaseWaffleMixin):
    """
    Checks that as switch is active, or 404. Operates like the FBV decorator
    waffle_switch.
    """

    waffle_switch: str | None = None

    def dispatch(self, request, *args, **kwargs):
        active = self.validate_waffle(self.waffle_switch, switch_is_active)

        if not active:
            return self.invalid_waffle()

        return super().dispatch(request, *args, **kwargs)



from unittest import skipUnless

from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.template import Template
from django.template.base import VariableNode
from django.test import RequestFactory

from test_app import views
from waffle.middleware import WaffleMiddleware
from waffle.tests.base import TestCase


def get():
    request = RequestFactory().get('/foo')
    request.user = AnonymousUser()
    return request


def get_response(request):
    return 'hello'


def process_request(request, view):
    response = view(request)
    return WaffleMiddleware(get_response).process_response(request, response)


class WaffleTemplateTests(TestCase):
    def test_django_tags(self):
        request = get()
        response = process_request(request, views.flag_in_django)
        self.assertContains(response, 'flag off')
        self.assertContains(response, 'switch off')
        self.assertContains(response, 'sample')
        self.assertContains(response, 'flag_var off')
        self.assertContains(response, 'switch_var off')
        self.assertContains(response, 'sample_var')
        self.assertContains(response, 'window.waffle =')

    def test_get_nodes_by_type(self):
        """WaffleNode.get_nodes_by_type() should find all child nodes."""
        test_template = Template(
            '{% load waffle_tags %}'
            '{% switch "x" %}'
            '{{ a }}'
            '{% else %}'
            '{{ b }}'
            '{% endswitch %}'
        )
        children = test_template.nodelist.get_nodes_by_type(VariableNode)
        self.assertEqual(len(children), 2)

    def test_no_request_context(self):
        """Switches and Samples shouldn't require a request context."""
        request = get()
        content = process_request(request, views.no_request_context)
        assert 'switch off' in content
        assert 'sample' in content

    @skipUnless(
        settings.JINJA_INSTALLED,
        "django-jinja is currently unmaintained and uncompatible with django >= 3.0"
    )
    def test_jinja_tags(self):
        request = get()
        response = process_request(request, views.flag_in_jinja)
        self.assertContains(response, 'flag off')
        self.assertContains(response, 'switch off')
        self.assertContains(response, 'sample')
        self.assertContains(response, 'window.waffle =')



from django import template
from django.template.base import VariableDoesNotExist

from waffle import flag_is_active, sample_is_active, switch_is_active
from waffle.views import _generate_waffle_js


register = template.Library()


class WaffleNode(template.Node):
    child_nodelists = ('nodelist_true', 'nodelist_false')

    def __init__(self, nodelist_true, nodelist_false, condition, name,
                 compiled_name):
        self.nodelist_true = nodelist_true
        self.nodelist_false = nodelist_false
        self.condition = condition
        self.name = name
        self.compiled_name = compiled_name

    def __repr__(self):
        return f'<Waffle node: {self.name}>'

    def __iter__(self):
        yield from self.nodelist_true
        yield from self.nodelist_false

    def render(self, context):
        try:
            name = self.compiled_name.resolve(context)
        except VariableDoesNotExist:
            name = self.name
        if not name:
            name = self.name
        if self.condition(context.get('request', None), name):
            return self.nodelist_true.render(context)
        return self.nodelist_false.render(context)

    @classmethod
    def handle_token(cls, parser, token, kind, condition):
        bits = token.split_contents()
        if len(bits) < 2:
            raise template.TemplateSyntaxError(f"{bits[0]!r} tag requires an argument")

        name = bits[1]
        compiled_name = parser.compile_filter(name)

        nodelist_true = parser.parse(('else', f'end{kind}'))
        token = parser.next_token()
        if token.contents == 'else':
            nodelist_false = parser.parse((f'end{kind}',))
            parser.delete_first_token()
        else:
            nodelist_false = template.NodeList()

        return cls(nodelist_true, nodelist_false, condition,
                   name, compiled_name)


@register.tag
def flag(parser, token):
    return WaffleNode.handle_token(parser, token, 'flag', flag_is_active)


@register.tag
def switch(parser, token):
    return WaffleNode.handle_token(parser, token, 'switch', lambda request, name: switch_is_active(name))


@register.tag
def sample(parser, token):
    return WaffleNode.handle_token(parser, token, 'sample', lambda request, name: sample_is_active(name))


class InlineWaffleJSNode(template.Node):
    def render(self, context):
        return _generate_waffle_js(context['request'])


@register.tag
def wafflejs(parser, token):
    return InlineWaffleJSNode()



from typing import Any

from django.core.management.base import BaseCommand, CommandError, CommandParser

from waffle import get_waffle_sample_model


class Command(BaseCommand):
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            'name',
            nargs='?',
            help='The name of the sample.')
        parser.add_argument(
            'percent',
            nargs='?',
            type=int,
            help='The percentage of the time this sample will be active.')
        parser.add_argument(
            '-l', '--list',
            action='store_true', dest='list_samples', default=False,
            help='List existing samples.')
        parser.add_argument(
            '--create',
            action='store_true',
            dest='create',
            default=False,
            help='If the sample does not exist, create it.')

    help = 'Change percentage of a sample.'

    def handle(self, *args: Any, **options: Any) -> None:
        if options['list_samples']:
            self.stdout.write('Samples:')
            for sample in get_waffle_sample_model().objects.iterator():
                self.stdout.write(f'{sample.name}: {sample.percent:.1f}%')
            self.stdout.write('')
            return

        sample_name = options['name']
        percent = options['percent']

        if not (sample_name and percent):
            raise CommandError(
                'You need to specify a sample name and percentage.'
            )

        try:
            percent = float(percent)
            if not (0.0 <= percent <= 100.0):
                raise ValueError
        except ValueError:
            raise CommandError('You need to enter a valid percentage value.')

        if options['create']:
            sample, created = get_waffle_sample_model().objects.get_or_create(
                name=sample_name, defaults={'percent': 0})
            if created:
                self.stdout.write(f'Creating sample: {sample_name}')
        else:
            try:
                sample = get_waffle_sample_model().objects.get(name=sample_name)
            except get_waffle_sample_model().DoesNotExist:
                raise CommandError('This sample does not exist.')

        sample.percent = percent
        sample.save()



COOKIE = 'dwf_%s'
TEST_COOKIE = 'dwft_%s'
SECURE = True
MAX_AGE = 2592000  # 1 month in seconds

CACHE_PREFIX = 'waffle:'
CACHE_NAME = 'default'
FLAG_CACHE_KEY = 'flag:%s'
FLAG_USERS_CACHE_KEY = 'flag:%s:users'
FLAG_GROUPS_CACHE_KEY = 'flag:%s:groups'
ALL_FLAGS_CACHE_KEY = 'flags:all'
SAMPLE_CACHE_KEY = 'sample:%s'
ALL_SAMPLES_CACHE_KEY = 'samples:all'
SWITCH_CACHE_KEY = 'switch:%s'
ALL_SWITCHES_CACHE_KEY = 'switches:all'

FLAG_DEFAULT = False
SAMPLE_DEFAULT = False
SWITCH_DEFAULT = False

READ_FROM_WRITE_DB = False

CREATE_MISSING_FLAGS = False
CREATE_MISSING_SAMPLES = False
CREATE_MISSING_SWITCHES = False

LOG_MISSING_FLAGS = None
LOG_MISSING_SAMPLES = None
LOG_MISSING_SWITCHES = None

OVERRIDE = False
ENABLE_ADMIN_PAGES = True



# Generated by Django 2.2 on 2022-07-17 19:34

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('test_app', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='CustomSample',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The human/computer readable name.', max_length=100, unique=True, verbose_name='Name')),
                ('percent', models.DecimalField(decimal_places=1, help_text='A number between 0.0 and 100.0 to indicate a percentage of the time this sample will be active.', max_digits=4, verbose_name='Percent')),
                ('note', models.TextField(blank=True, help_text='Note where this Sample is used.', verbose_name='Note')),
                ('created', models.DateTimeField(db_index=True, default=django.utils.timezone.now, help_text='Date when this Sample was created.', verbose_name='Created')),
                ('modified', models.DateTimeField(default=django.utils.timezone.now, help_text='Date when this Sample was last modified.', verbose_name='Modified')),
            ],
            options={
                'verbose_name': 'Sample',
                'verbose_name_plural': 'Samples',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='CustomSwitch',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The human/computer readable name.', max_length=100, unique=True, verbose_name='Name')),
                ('active', models.BooleanField(default=False, help_text='Is this switch active?', verbose_name='Active')),
                ('note', models.TextField(blank=True, help_text='Note where this Switch is used.', verbose_name='Note')),
                ('created', models.DateTimeField(db_index=True, default=django.utils.timezone.now, help_text='Date when this Switch was created.', verbose_name='Created')),
                ('modified', models.DateTimeField(default=django.utils.timezone.now, help_text='Date when this Switch was last modified.', verbose_name='Modified')),
            ],
            options={
                'verbose_name': 'Switch',
                'verbose_name_plural': 'Switches',
                'abstract': False,
            },
        ),
        migrations.AlterField(
            model_name='companyawareflag',
            name='everyone',
            field=models.BooleanField(blank=True, help_text='Flip this flag on (Yes) or off (No) for everyone, overriding all other settings. Leave as Unknown to use normally.', null=True, verbose_name='Everyone'),
        ),
    ]



# Generated by Django 2.0 on 2018-03-05 22:21

from django.conf import settings
from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('waffle', '0002_auto_20161201_0958'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='flag',
            options={'verbose_name': 'Flag', 'verbose_name_plural': 'Flags'},
        ),
        migrations.AlterModelOptions(
            name='sample',
            options={'verbose_name': 'Sample', 'verbose_name_plural': 'Samples'},
        ),
        migrations.AlterModelOptions(
            name='switch',
            options={'verbose_name': 'Switch', 'verbose_name_plural': 'Switches'},
        ),
        migrations.AlterField(
            model_name='flag',
            name='authenticated',
            field=models.BooleanField(default=False, help_text='Flag always active for authenticated users?', verbose_name='Authenticated'),
        ),
        migrations.AlterField(
            model_name='flag',
            name='created',
            field=models.DateTimeField(db_index=True, default=django.utils.timezone.now, help_text='Date when this Flag was created.', verbose_name='Created'),
        ),
        migrations.AlterField(
            model_name='flag',
            name='everyone',
            field=models.NullBooleanField(help_text='Flip this flag on (Yes) or off (No) for everyone, overriding all other settings. Leave as Unknown to use normally.', verbose_name='Everyone'),
        ),
        migrations.AlterField(
            model_name='flag',
            name='groups',
            field=models.ManyToManyField(blank=True, help_text='Activate this flag for these user groups.', to='auth.Group', verbose_name='Groups'),
        ),
        migrations.AlterField(
            model_name='flag',
            name='languages',
            field=models.TextField(blank=True, default='', help_text='Activate this flag for users with one of these languages (comma-separated list)', verbose_name='Languages'),
        ),
        migrations.AlterField(
            model_name='flag',
            name='modified',
            field=models.DateTimeField(default=django.utils.timezone.now, help_text='Date when this Flag was last modified.', verbose_name='Modified'),
        ),
        migrations.AlterField(
            model_name='flag',
            name='name',
            field=models.CharField(help_text='The human/computer readable name.', max_length=100, unique=True, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='flag',
            name='note',
            field=models.TextField(blank=True, help_text='Note where this Flag is used.', verbose_name='Note'),
        ),
        migrations.AlterField(
            model_name='flag',
            name='percent',
            field=models.DecimalField(blank=True, decimal_places=1, help_text='A number between 0.0 and 99.9 to indicate a percentage of users for whom this flag will be active.', max_digits=3, null=True, verbose_name='Percent'),
        ),
        migrations.AlterField(
            model_name='flag',
            name='rollout',
            field=models.BooleanField(default=False, help_text='Activate roll-out mode?', verbose_name='Rollout'),
        ),
        migrations.AlterField(
            model_name='flag',
            name='staff',
            field=models.BooleanField(default=False, help_text='Flag always active for staff?', verbose_name='Staff'),
        ),
        migrations.AlterField(
            model_name='flag',
            name='superusers',
            field=models.BooleanField(default=True, help_text='Flag always active for superusers?', verbose_name='Superusers'),
        ),
        migrations.AlterField(
            model_name='flag',
            name='testing',
            field=models.BooleanField(default=False, help_text='Allow this flag to be set for a session for user testing', verbose_name='Testing'),
        ),
        migrations.AlterField(
            model_name='flag',
            name='users',
            field=models.ManyToManyField(blank=True, help_text='Activate this flag for these users.', to=settings.AUTH_USER_MODEL, verbose_name='Users'),
        ),
        migrations.AlterField(
            model_name='sample',
            name='created',
            field=models.DateTimeField(db_index=True, default=django.utils.timezone.now, help_text='Date when this Sample was created.', verbose_name='Created'),
        ),
        migrations.AlterField(
            model_name='sample',
            name='modified',
            field=models.DateTimeField(default=django.utils.timezone.now, help_text='Date when this Sample was last modified.', verbose_name='Modified'),
        ),
        migrations.AlterField(
            model_name='sample',
            name='name',
            field=models.CharField(help_text='The human/computer readable name.', max_length=100, unique=True, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='sample',
            name='note',
            field=models.TextField(blank=True, help_text='Note where this Sample is used.', verbose_name='Note'),
        ),
        migrations.AlterField(
            model_name='sample',
            name='percent',
            field=models.DecimalField(decimal_places=1, help_text='A number between 0.0 and 100.0 to indicate a percentage of the time this sample will be active.', max_digits=4, verbose_name='Percent'),
        ),
        migrations.AlterField(
            model_name='switch',
            name='active',
            field=models.BooleanField(default=False, help_text='Is this switch active?', verbose_name='Active'),
        ),
        migrations.AlterField(
            model_name='switch',
            name='created',
            field=models.DateTimeField(db_index=True, default=django.utils.timezone.now, help_text='Date when this Switch was created.', verbose_name='Created'),
        ),
        migrations.AlterField(
            model_name='switch',
            name='modified',
            field=models.DateTimeField(default=django.utils.timezone.now, help_text='Date when this Switch was last modified.', verbose_name='Modified'),
        ),
        migrations.AlterField(
            model_name='switch',
            name='name',
            field=models.CharField(help_text='The human/computer readable name.', max_length=100, unique=True, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='switch',
            name='note',
            field=models.TextField(blank=True, help_text='Note where this Switch is used.', verbose_name='Note'),
        ),
    ]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest

from waffle.utils import get_setting
from django.apps import apps as django_apps

if TYPE_CHECKING:
    from waffle.models import AbstractBaseFlag, AbstractBaseSample, AbstractBaseSwitch

__version__ = '5.0.0'


def flag_is_active(request: HttpRequest, flag_name: str, read_only: bool = False) -> bool | None:
    flag = get_waffle_flag_model().get(flag_name)
    return flag.is_active(request, read_only=read_only)


def switch_is_active(switch_name: str) -> bool:
    switch = get_waffle_switch_model().get(switch_name)
    return switch.is_active()


def sample_is_active(sample_name: str) -> bool:
    sample = get_waffle_sample_model().get(sample_name)
    return sample.is_active()


def get_waffle_flag_model() -> type[AbstractBaseFlag]:
    return get_waffle_model('FLAG_MODEL')


def get_waffle_switch_model() -> type[AbstractBaseSwitch]:
    return get_waffle_model('SWITCH_MODEL')


def get_waffle_sample_model() -> type[AbstractBaseSample]:
    return get_waffle_model('SAMPLE_MODEL')


def get_waffle_model(setting_name: str) -> (
    type[AbstractBaseFlag | AbstractBaseSwitch | AbstractBaseSample]
):
    """
    Returns the waffle Flag model that is active in this project.
    """
    default_model = {
        'FLAG_MODEL': 'waffle.Flag',
        'SWITCH_MODEL': 'waffle.Switch',
        'SAMPLE_MODEL': 'waffle.Sample',
    }

    # Add backwards compatibility by not requiring adding of model setting
    # for everyone who upgrades.  At some point it would be helpful to
    # require this to be defined explicitly, but no for now, to remove
    # pain from upgrading.
    default = default_model[setting_name]
    flag_model_name = get_setting(setting_name, default)

    try:
        return django_apps.get_model(flag_model_name)
    except ValueError:
        raise ImproperlyConfigured(f"WAFFLE_{setting_name} must be of the form 'app_label.model_name'")
    except LookupError:
        raise ImproperlyConfigured(
            f"WAFFLE_{setting_name} refers to model '{flag_model_name}' that has not been installed"
        )



from __future__ import annotations

from typing import Any

from django.contrib import admin
from django.contrib.admin.models import LogEntry, CHANGE, DELETION
from django.contrib.admin.widgets import ManyToManyRawIdWidget
from django.contrib.contenttypes.models import ContentType
from django.http import HttpRequest
from django.utils.html import escape
from django.utils.translation import gettext_lazy as _

from waffle.models import Flag, Sample, Switch
from waffle.utils import get_setting


class BaseAdmin(admin.ModelAdmin):
    search_fields = ('name', 'note')

    def get_actions(self, request: HttpRequest) -> dict[str, Any]:
        actions = super().get_actions(request)
        if 'delete_selected' in actions:
            del actions['delete_selected']
        return actions


def _add_log_entry(user, model, description, action_flag):
    LogEntry.objects.create(
        user=user,
        content_type=ContentType.objects.get_for_model(type(model)),
        object_id=model.id,
        object_repr=model.name + " " + description,
        action_flag=action_flag
    )


@admin.action(
    description=_('Enable selected flags for everyone'),
    permissions=('change',),
)
def enable_for_all(ma, request, qs):
    # Iterate over all objects to cause cache invalidation.
    for f in qs.all():
        _add_log_entry(request.user, f, "on", CHANGE)
        f.everyone = True
        f.save()


@admin.action(
    description=_('Disable selected flags for everyone'),
    permissions=('change',),
)
def disable_for_all(ma, request, qs):
    # Iterate over all objects to cause cache invalidation.
    for f in qs.all():
        _add_log_entry(request.user, f, "off", CHANGE)
        f.everyone = False
        f.save()


@admin.action(
    description=_('Delete selected'),
    permissions=('delete',),
)
def delete_individually(ma, request, qs):
    # Iterate over all objects to cause cache invalidation.
    for f in qs.all():
        _add_log_entry(request.user, f, "deleted", DELETION)
        f.delete()


class InformativeManyToManyRawIdWidget(ManyToManyRawIdWidget):
    """Widget for ManyToManyField to Users.

    Will display the names of the users in a parenthesised list after the
    input field. This widget works with all models that have a "name" field.
    """
    def label_and_url_for_value(self, values: Any) -> tuple[str, str]:
        names = []
        key = self.rel.get_related_field().name
        for value in values:
            try:
                name = self.rel.model._default_manager \
                    .using(self.db) \
                    .get(**{key: value})
                names.append(escape(str(name)))
            except self.rel.model.DoesNotExist:
                names.append('<missing>')
        return "(" + ", ".join(names) + ")", ""


class FlagAdmin(BaseAdmin):
    actions = [enable_for_all, disable_for_all, delete_individually]
    list_display = ('name', 'note', 'everyone', 'percent', 'superusers',
                    'staff', 'authenticated', 'languages')
    list_filter = ('everyone', 'superusers', 'staff', 'authenticated')
    raw_id_fields = ('users', )
    ordering = ('-id',)

    def formfield_for_dbfield(self, db_field, **kwargs):
        if db_field.name == 'users':
            kwargs.pop('request', None)
            kwargs['widget'] = \
                InformativeManyToManyRawIdWidget(db_field.remote_field,
                                                 self.admin_site,
                                                 using=kwargs.get("using"))
            return db_field.formfield(**kwargs)
        return super().formfield_for_dbfield(db_field, **kwargs)


@admin.action(
    description=_('Enable selected switches'),
    permissions=('change',),
)
def enable_switches(ma, request, qs):
    for switch in qs:
        _add_log_entry(request.user, switch, "on", CHANGE)
        switch.active = True
        switch.save()


@admin.action(
    description=_('Disable selected switches'),
    permissions=('change',),
)
def disable_switches(ma, request, qs):
    for switch in qs:
        _add_log_entry(request.user, switch, "off", CHANGE)
        switch.active = False
        switch.save()


class SwitchAdmin(BaseAdmin):
    actions = [enable_switches, disable_switches, delete_individually]
    list_display = ('name', 'active', 'note', 'created', 'modified')
    list_filter = ('active',)
    ordering = ('-id',)


class SampleAdmin(BaseAdmin):
    actions = [delete_individually]
    list_display = ('name', 'percent', 'note', 'created', 'modified')
    ordering = ('-id',)


if get_setting('ENABLE_ADMIN_PAGES'):
    admin.site.register(Flag, FlagAdmin)
    admin.site.register(Sample, SampleAdmin)
    admin.site.register(Switch, SwitchAdmin)



import logging
import random
import threading
import unittest

from django.contrib.auth import get_user_model
from django.conf import settings
from django.contrib.auth.models import AnonymousUser, Group
from django.db import connection, transaction
from django.test import RequestFactory, TransactionTestCase

from django.test.utils import override_settings

from unittest import mock

import waffle
from test_app import views
from test_app.models import CompanyAwareFlag, CustomSample, CustomSwitch, Company
from waffle.middleware import WaffleMiddleware
from waffle.tests.base import TestCase


DATABASES = {'default', 'readonly'}


def get(**kw):
    request = RequestFactory().get('/foo', data=kw)
    request.user = AnonymousUser()
    return request


def get_response(request):
    return 'hello'


def process_request(request, view):
    response = view(request)
    return WaffleMiddleware(get_response).process_response(request, response)


class WaffleTests(TestCase):
    databases = DATABASES

    def assert_flag_dynamically_created_with_value(self, expected_value):
        FLAG_NAME = 'my_dynamically_created_flag'
        flag_model = waffle.get_waffle_flag_model()

        assert flag_model.objects.count() == 0
        assert expected_value == waffle.flag_is_active(get(), FLAG_NAME)
        assert flag_model.objects.count() == 1

        flag = flag_model.objects.get(name=FLAG_NAME)
        assert flag.name == FLAG_NAME

        # We assert no queries are made to ensure flags created when the
        # `CREATE_MISSING_FLAGS` setting is active are properly cached.
        with self.assertNumQueries(0):
            assert expected_value == waffle.flag_is_active(get(), FLAG_NAME)

    def test_persist_active_flag(self):
        waffle.get_waffle_flag_model().objects.create(name='myflag', percent='0.1')
        request = get()

        # Flag stays on.
        request.COOKIES['dwf_myflag'] = 'True'
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'on', response.content)
        assert 'dwf_myflag' in response.cookies
        self.assertEqual('True', response.cookies['dwf_myflag'].value)

    def test_persist_inactive_flag(self):
        waffle.get_waffle_flag_model().objects.create(name='myflag', percent='99.9')
        request = get()

        # Flag stays off.
        request.COOKIES['dwf_myflag'] = 'False'
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)
        assert 'dwf_myflag' in response.cookies
        self.assertEqual('False', response.cookies['dwf_myflag'].value)

    def test_no_set_unused_flag(self):
        """An unused flag shouldn't have its cookie reset."""
        request = get()
        request.COOKIES['dwf_unused'] = 'True'
        response = process_request(request, views.flag_in_view)
        assert 'dwf_unused' not in response.cookies

    def test_superuser(self):
        """Test the superuser switch."""
        waffle.get_waffle_flag_model().objects.create(name='myflag', superusers=True)
        request = get()
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)
        assert 'dwf_myflag' not in response.cookies

        superuser = get_user_model()(username='foo', is_superuser=True)
        request.user = superuser
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'on', response.content)
        assert 'dwf_myflag' not in response.cookies

        non_superuser = get_user_model()(username='bar', is_superuser=False)
        non_superuser.save()
        request.user = non_superuser
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)
        assert 'dwf_myflag' not in response.cookies

    def test_staff(self):
        """Test the staff switch."""
        waffle.get_waffle_flag_model().objects.create(name='myflag', staff=True)
        request = get()
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)
        assert 'dwf_myflag' not in response.cookies

        staff = get_user_model()(username='foo', is_staff=True)
        request.user = staff
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'on', response.content)
        assert 'dwf_myflag' not in response.cookies

        non_staff = get_user_model()(username='foo', is_staff=False)
        non_staff.save()
        request.user = non_staff
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)
        assert 'dwf_myflag' not in response.cookies

    def test_languages(self):
        waffle.get_waffle_flag_model().objects.create(name='myflag', languages='en,fr')
        request = get()
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)

        request.LANGUAGE_CODE = 'en'
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'on', response.content)

        request.LANGUAGE_CODE = 'de'
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)

    def test_user(self):
        """Test the per-user switch."""
        user = get_user_model().objects.create(username='foo')
        flag = waffle.get_waffle_flag_model().objects.create(name='myflag')
        flag.users.add(user)

        request = get()
        request.user = user
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'on', response.content)
        assert 'dwf_myflag' not in response.cookies

        request.user = get_user_model().objects.create(username='someone_else')
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)
        assert 'dwf_myflag' not in response.cookies

        # Unsetting the flag on a user should have an effect.
        flag.users.remove(user)
        request.user = user
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)

    def test_remove_from_user(self):
        """Same operation of `test_user` but performed with reverse relation"""
        user = get_user_model().objects.create(username='foo')
        flag = waffle.get_waffle_flag_model().objects.create(name='myflag')
        flag.users.add(user)

        request = get()
        request.user = user
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'on', response.content)
        assert 'dwf_myflag' not in response.cookies

        request.user = get_user_model().objects.create(username='someone_else')
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)
        assert 'dwf_myflag' not in response.cookies

        # Unsetting the flag on a user should have an effect.
        user.flag_set.remove(flag)
        request.user = user
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)

    def test_group(self):
        """Test the per-group switch."""
        group = Group.objects.create(name='foo')
        user = get_user_model().objects.create(username='bar')
        user.groups.add(group)

        flag = waffle.get_waffle_flag_model().objects.create(name='myflag')
        flag.groups.add(group)

        request = get()
        request.user = user
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'on', response.content)
        assert 'dwf_myflag' not in response.cookies

        request.user = get_user_model()(username='someone_else')
        request.user.save()
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)
        assert 'dwf_myflag' not in response.cookies

        # Unsetting the flag on a group should have an effect.
        flag.groups.remove(group)
        request.user = user
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)

    def test_remove_from_group(self):
        """Same operation of `test_group` but performed with reverse relation"""
        group = Group.objects.create(name='foo')
        user = get_user_model().objects.create(username='bar')
        user.groups.add(group)

        flag = waffle.get_waffle_flag_model().objects.create(name='myflag')
        flag.groups.add(group)

        request = get()
        request.user = user
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'on', response.content)
        assert 'dwf_myflag' not in response.cookies

        request.user = get_user_model()(username='someone_else')
        request.user.save()
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)
        assert 'dwf_myflag' not in response.cookies

        # Unsetting the flag on a group should have an effect.
        group.flag_set.remove(flag)
        request.user = user
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)

    def test_authenticated(self):
        """Test the authenticated/anonymous switch."""
        waffle.get_waffle_flag_model().objects.create(name='myflag', authenticated=True)

        request = get()
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)
        assert 'dwf_myflag' not in response.cookies

        request.user = get_user_model()(username='foo')
        assert request.user.is_authenticated
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'on', response.content)
        assert 'dwf_myflag' not in response.cookies

    def test_everyone_on(self):
        """Test the 'everyone' switch on."""
        waffle.get_waffle_flag_model().objects.create(name='myflag', everyone=True)

        request = get()
        request.COOKIES['dwf_myflag'] = 'False'
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'on', response.content)
        assert 'dwf_myflag' not in response.cookies

        request.user = get_user_model()(username='foo')
        assert request.user.is_authenticated
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'on', response.content)
        assert 'dwf_myflag' not in response.cookies

    def test_everyone_off(self):
        """Test the 'everyone' switch off."""
        waffle.get_waffle_flag_model().objects.create(
            name="myflag", everyone=False, authenticated=True
        )

        request = get()
        request.COOKIES['dwf_myflag'] = 'True'
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)
        assert 'dwf_myflag' not in response.cookies

        request.user = get_user_model()(username='foo')
        assert request.user.is_authenticated
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)
        assert 'dwf_myflag' not in response.cookies

    def test_percent(self):
        """If you have no cookie, you get a cookie!"""
        waffle.get_waffle_flag_model().objects.create(name='myflag', percent='50.0')
        request = get()
        response = process_request(request, views.flag_in_view)
        assert 'dwf_myflag' in response.cookies

    @mock.patch.object(random, 'uniform')
    def test_reroll(self, uniform):
        """Even without a cookie, calling flag_is_active twice should return
        the same value."""
        waffle.get_waffle_flag_model().objects.create(name='myflag', percent='50.0')
        # Make sure we're not really random.
        request = get()  # Create a clean request.
        assert not hasattr(request, 'waffles')
        uniform.return_value = '10'  # < 50. Flag is True.
        assert waffle.flag_is_active(request, 'myflag')
        assert hasattr(request, 'waffles')  # We should record this flag.
        assert 'myflag' in request.waffles
        assert request.waffles['myflag'][0]
        uniform.return_value = '70'  # > 50. Normally, Flag would be False.
        assert waffle.flag_is_active(request, 'myflag')
        assert request.waffles['myflag'][0]

    def test_undefined(self):
        """Undefined flags are always false."""
        request = get()
        assert not waffle.flag_is_active(request, 'foo')

    @override_settings(WAFFLE_FLAG_DEFAULT=True)
    def test_undefined_default(self):
        """WAFFLE_FLAG_DEFAULT controls undefined flags."""
        request = get()
        assert waffle.flag_is_active(request, 'foo')

    @override_settings(WAFFLE_OVERRIDE=True)
    def test_override(self):
        request = get(foo='1')
        waffle.get_waffle_flag_model().objects.create(name='foo')  # Off for everyone.
        assert waffle.flag_is_active(request, 'foo')

    def test_testing_flag(self):
        waffle.get_waffle_flag_model().objects.create(name='foo', testing=True)
        request = get(dwft_foo='1')
        assert waffle.flag_is_active(request, 'foo')
        assert 'foo' in request.waffle_tests
        assert request.waffle_tests['foo']

        # GET param should override cookie
        request = get(dwft_foo='0')
        request.COOKIES['dwft_foo'] = 'True'
        assert not waffle.flag_is_active(request, 'foo')
        assert 'foo' in request.waffle_tests
        assert not request.waffle_tests['foo']

    def test_testing_disabled_flag(self):
        waffle.get_waffle_flag_model().objects.create(name='foo')
        request = get(dwft_foo='1')
        assert not waffle.flag_is_active(request, 'foo')
        assert not hasattr(request, 'waffle_tests')

        request = get(dwft_foo='0')
        assert not waffle.flag_is_active(request, 'foo')
        assert not hasattr(request, 'waffle_tests')

    def test_testing_flag_header(self):
        waffle.get_waffle_flag_model().objects.create(name='foo', testing=True)
        request = RequestFactory().get('/foo', HTTP_DWFT_FOO='1')
        request.user = AnonymousUser()
        assert waffle.flag_is_active(request, 'foo')
        assert 'foo' in request.waffle_tests
        assert request.waffle_tests['foo']

        # header should override cookie
        request = RequestFactory().get('/foo', HTTP_DWFT_FOO='0')
        request.user = AnonymousUser()
        request.COOKIES['dwft_foo'] = 'True'
        assert not waffle.flag_is_active(request, 'foo')
        assert 'foo' in request.waffle_tests
        assert not request.waffle_tests['foo']

    def test_set_then_unset_testing_flag(self):
        waffle.get_waffle_flag_model().objects.create(name='myflag', testing=True)
        response = self.client.get('/flag_in_view?dwft_myflag=1')
        self.assertEqual(b'on', response.content)

        response = self.client.get('/flag_in_view')
        self.assertEqual(b'on', response.content)

        response = self.client.get('/flag_in_view?dwft_myflag=0')
        self.assertEqual(b'off', response.content)

        response = self.client.get('/flag_in_view')
        self.assertEqual(b'off', response.content)

        response = self.client.get('/flag_in_view?dwft_myflag=1')
        self.assertEqual(b'on', response.content)

    @override_settings(DATABASE_ROUTERS=['waffle.tests.base.ReplicationRouter'])
    def test_everyone_on_read_from_write_db(self):
        flag = waffle.get_waffle_flag_model().objects.create(name='myflag', everyone=True)

        request = get()
        response = process_request(request, views.flag_in_view)
        # By default, flag_is_active should hit whatever it configured as the
        # read DB (so values will be stale if replication is lagged).
        self.assertEqual(b'off', response.content)

        with override_settings(WAFFLE_READ_FROM_WRITE_DB=True):
            # Save the flag again to flush the cache.
            flag.save()

            # The next read should now be directed to the write DB, ensuring
            # the cache and DB are in sync.
            response = process_request(request, views.flag_in_view)
            self.assertEqual(b'on', response.content)

    @override_settings(
        WAFFLE_FLAG_MODEL="test_app.CompanyAwareFlag",
        AUTH_USER_MODEL="test_app.CompanyUser",
    )
    def test_pluggable_flag_model(self):
        flag_model = waffle.get_waffle_flag_model()
        self.assertEqual(CompanyAwareFlag, flag_model)

        acme_company = Company.objects.create(name='Acme Ltd.')
        feline_company = Company.objects.create(name='Feline LLC')

        acme_company_flag = waffle.get_waffle_flag_model().objects.create(
            name="myflag", superusers=True
        )
        acme_company_flag.companies.add(acme_company)

        request = get()
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)
        assert 'dwf_myflag' not in response.cookies

        acme_user = get_user_model()(username='acme.mcfield', company=acme_company)
        request.user = acme_user
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'on', response.content)
        assert 'dwf_myflag' not in response.cookies

        feline_user = get_user_model()(username='acme.mcfield', company=feline_company)
        request.user = feline_user
        response = process_request(request, views.flag_in_view)
        self.assertEqual(b'off', response.content)
        assert 'dwf_myflag' not in response.cookies

    @override_settings(WAFFLE_CREATE_MISSING_FLAGS=True)
    @override_settings(WAFFLE_FLAG_DEFAULT=False)
    def test_flag_created_dynamically_default_false(self):
        self.assert_flag_dynamically_created_with_value(False)

    @override_settings(WAFFLE_CREATE_MISSING_FLAGS=True)
    @override_settings(WAFFLE_FLAG_DEFAULT=True)
    def test_flag_created_dynamically_default_true(self):
        self.assert_flag_dynamically_created_with_value(True)

    @mock.patch('waffle.models.logger')
    def test_no_logging_missing_flag_by_default(self, mock_logger):
        request = get()
        waffle.flag_is_active(request, 'foo')
        mock_logger.log.call_count == 0

    @override_settings(WAFFLE_LOG_MISSING_FLAGS=logging.WARNING)
    @mock.patch('waffle.models.logger')
    def test_logging_missing_flag(self, mock_logger):
        request = get()
        waffle.flag_is_active(request, 'foo')
        mock_logger.log.assert_called_with(logging.WARNING, 'Flag %s not found', 'foo')


class SwitchTests(TestCase):
    databases = DATABASES

    def assert_switch_dynamically_created_with_value(self, expected_value):
        SWITCH_NAME = 'my_dynamically_created_switch'

        assert waffle.get_waffle_switch_model().objects.count() == 0
        assert expected_value == waffle.switch_is_active(SWITCH_NAME)
        assert waffle.get_waffle_switch_model().objects.count() == 1

        switch = waffle.get_waffle_switch_model().objects.get(name=SWITCH_NAME)
        assert switch.name == SWITCH_NAME
        assert expected_value == switch.active

        # We assert no queries are made to ensure switches created when the
        # `CREATE_MISSING_SWITCHES` setting is active are properly cached.
        with self.assertNumQueries(0):
            assert expected_value == waffle.switch_is_active(SWITCH_NAME)

    def test_switch_active(self):
        switch = waffle.get_waffle_switch_model().objects.create(
            name="myswitch", active=True
        )
        assert waffle.switch_is_active(switch.name)

    def test_switch_inactive(self):
        switch = waffle.get_waffle_switch_model().objects.create(
            name="myswitch", active=False
        )
        assert not waffle.switch_is_active(switch.name)

    def test_switch_active_from_cache(self):
        """Do not make two queries for an existing active switch."""
        switch = waffle.get_waffle_switch_model().objects.create(
            name="myswitch", active=True
        )
        # Get the value once so that it will be put into the cache
        assert waffle.switch_is_active(switch.name)
        queries = len(connection.queries)
        assert waffle.switch_is_active(switch.name)
        self.assertEqual(queries, len(connection.queries))

    def test_switch_inactive_from_cache(self):
        """Do not make two queries for an existing inactive switch."""
        switch = waffle.get_waffle_switch_model().objects.create(
            name="myswitch", active=False
        )
        # Get the value once so that it will be put into the cache
        assert not waffle.switch_is_active(switch.name)
        queries = len(connection.queries)
        assert not waffle.switch_is_active(switch.name)
        self.assertEqual(queries, len(connection.queries))

    def test_undefined(self):
        assert not waffle.switch_is_active('foo')

    @override_settings(WAFFLE_SWITCH_DEFAULT=True)
    def test_undefined_default(self):
        assert waffle.switch_is_active('foo')

    @override_settings(DEBUG=True)
    def test_no_query(self):
        """Do not make two queries for a non-existent switch."""
        assert not waffle.get_waffle_switch_model().objects.filter(name='foo').exists()
        queries = len(connection.queries)
        assert not waffle.switch_is_active('foo')
        assert len(connection.queries) > queries
        queries = len(connection.queries)
        assert not waffle.switch_is_active('foo')
        self.assertEqual(queries, len(connection.queries))

    @override_settings(DATABASE_ROUTERS=['waffle.tests.base.ReplicationRouter'])
    def test_read_from_write_db(self):
        switch = waffle.get_waffle_switch_model().objects.create(
            name="switch", active=True
        )

        # By default, switch_is_active should hit whatever it configured as the
        # read DB (so values will be stale if replication is lagged).
        assert not waffle.switch_is_active(switch.name)

        with override_settings(WAFFLE_READ_FROM_WRITE_DB=True):
            # Save the switch again to flush the cache.
            switch.save()

            # The next read should now be directed to the write DB, ensuring
            # the cache and DB are in sync.
            assert waffle.switch_is_active(switch.name)

    @override_settings(WAFFLE_SWITCH_MODEL='test_app.CustomSwitch')
    def test_pluggable_switch_model(self):
        switch_model = waffle.get_waffle_model('SWITCH_MODEL')
        self.assertEqual(CustomSwitch, switch_model)

        switch_model.objects.create(name='test_switch_off', active=False)
        switch_model.objects.create(name='test_switch_on', active=True)

        assert not waffle.switch_is_active('test_switch_off')
        assert waffle.switch_is_active('test_switch_on')

    @override_settings(WAFFLE_CREATE_MISSING_SWITCHES=True)
    @override_settings(WAFFLE_SWITCH_DEFAULT=False)
    def test_switch_created_dynamically_false(self):
        self.assert_switch_dynamically_created_with_value(False)

    @override_settings(WAFFLE_CREATE_MISSING_SWITCHES=True)
    @override_settings(WAFFLE_SWITCH_DEFAULT=True)
    def test_switch_created_dynamically_true(self):
        self.assert_switch_dynamically_created_with_value(True)

    @mock.patch('waffle.models.logger')
    def test_no_logging_missing_switch_by_default(self, mock_logger):
        waffle.switch_is_active('foo')
        mock_logger.log.call_count == 0

    @override_settings(WAFFLE_LOG_MISSING_SWITCHES=logging.WARNING)
    @mock.patch('waffle.models.logger')
    def test_logging_missing_switch(self, mock_logger):
        waffle.switch_is_active("foo")
        mock_logger.log.assert_called_with(
            logging.WARNING, "Switch %s not found", "foo"
        )


class SampleTests(TestCase):
    databases = DATABASES

    def assert_sample_dynamically_created_with_value(self, is_active, expected_value):
        SAMPLE_NAME = 'my_dynamically_created_sample'

        assert waffle.get_waffle_sample_model().objects.count() == 0
        assert is_active == waffle.sample_is_active(SAMPLE_NAME)
        assert waffle.get_waffle_sample_model().objects.count() == 1

        sample = waffle.get_waffle_sample_model().objects.get(name=SAMPLE_NAME)

        assert sample.name == SAMPLE_NAME
        assert sample.percent == expected_value

        # We assert no queries are made to ensure samples created when the
        # `CREATE_MISSING_SAMPLES` setting is active are properly cached.
        with self.assertNumQueries(0):
            assert is_active == waffle.sample_is_active(SAMPLE_NAME)

    def test_sample_100(self):
        sample = waffle.get_waffle_sample_model().objects.create(
            name="sample", percent="100.0"
        )
        assert waffle.sample_is_active(sample.name)

    def test_sample_0(self):
        sample = waffle.get_waffle_sample_model().objects.create(
            name="sample", percent="0.0"
        )
        assert not waffle.sample_is_active(sample.name)

    def test_undefined(self):
        assert not waffle.sample_is_active('foo')

    @override_settings(WAFFLE_SAMPLE_DEFAULT=True)
    def test_undefined_default(self):
        assert waffle.sample_is_active('foo')

    @override_settings(DATABASE_ROUTERS=['waffle.tests.base.ReplicationRouter'])
    def test_read_from_write_db(self):
        sample = waffle.get_waffle_sample_model().objects.create(
            name="sample", percent="100.0"
        )

        # By default, sample_is_active should hit whatever it configured as the
        # read DB (so values will be stale if replication is lagged).
        assert not waffle.sample_is_active(sample.name)

        with override_settings(WAFFLE_READ_FROM_WRITE_DB=True):
            # Save the sample again to flush the cache.
            sample.save()

            # The next read should now be directed to the write DB, ensuring
            # the cache and DB are in sync.
            assert waffle.sample_is_active(sample.name)

    @override_settings(WAFFLE_CREATE_MISSING_SAMPLES=True)
    @override_settings(WAFFLE_SAMPLE_DEFAULT=False)
    def test_sample_created_dynamically_default_false(self):
        self.assert_sample_dynamically_created_with_value(False, 0.0)

    @override_settings(WAFFLE_CREATE_MISSING_SAMPLES=True)
    @override_settings(WAFFLE_SAMPLE_DEFAULT=True)
    def test_sample_created_dynamically_default_true(self):
        self.assert_sample_dynamically_created_with_value(True, 100.0)

    @mock.patch('waffle.models.logger')
    def test_no_logging_missing_sample_by_default(self, mock_logger):
        waffle.switch_is_active('foo')
        mock_logger.log.call_count == 0

    @override_settings(WAFFLE_LOG_MISSING_SAMPLES=logging.WARNING)
    @mock.patch('waffle.models.logger')
    def test_logging_missing_sample(self, mock_logger):
        waffle.sample_is_active('foo')
        mock_logger.log.assert_called_with(logging.WARNING, 'Sample %s not found', 'foo')

    @override_settings(WAFFLE_SAMPLE_MODEL='test_app.CustomSample')
    def test_pluggable_sample_model(self):
        sample_model = waffle.get_waffle_model('SAMPLE_MODEL')
        self.assertEqual(CustomSample, sample_model)

        sample_model.objects.create(name='test_sample_off', percent=0)
        sample_model.objects.create(name='test_sample_on', percent=100)

        assert not waffle.sample_is_active('test_sample_off')
        assert waffle.sample_is_active('test_sample_on')


class TransactionTestMixin:
    """Mixin providing an abstract test case for writing in a transaction.
    """
    def create_toggle(self):
        """Create an inactive feature toggle (i.e. flag, switch, sample)."""
        raise NotImplementedError

    def flip_toggle(self, toggle):
        """Flip the toggle to an active state."""
        raise NotImplementedError

    def toggle_is_active(self, toggle):
        """Use the built-in *_is_active helper to check the toggle's state."""
        raise NotImplementedError

    @unittest.skipIf('sqlite3' in settings.DATABASES['default']['ENGINE'],
                     'This test uses threads, which the sqlite3 DB engine '
                     'does not support.')
    def test_flip_toggle_in_transaction(self):
        """Wait to invalidate the cache until after the current transaction.

        This test covers a potential race condition where, if the cache were
        flushed in the middle of a transaction, the next read from the database
        (before the transaction is committed) would get a stale value and cache
        it. See #296 for more context.
        """
        toggle = self.create_toggle()
        self.addCleanup(toggle.delete)

        written_in_background_thread = threading.Event()
        read_in_main_thread = threading.Event()

        @transaction.atomic
        def update_toggle():
            self.flip_toggle(toggle)

            # Signal to the main thread that the toggle has been updated, but
            # the transaction is not yet committed.
            written_in_background_thread.set()

            # Pause here to allow the main thread to make an assertion.
            read_in_main_thread.wait(timeout=1)

        # Start a background thread to update the toggle in a transaction.
        t = threading.Thread(target=update_toggle)
        t.daemon = True
        t.start()

        # After the toggle is updated but before the transaction is committed,
        # the cache will still have the previous value.
        written_in_background_thread.wait(timeout=1)
        assert not self.toggle_is_active(toggle)

        # After the transaction is committed, the cache should have been
        # invalidated, hence the next call to *_is_active should have the
        # correct value.
        read_in_main_thread.set()
        t.join(timeout=1)
        assert self.toggle_is_active(toggle)


class FlagTransactionTests(TransactionTestMixin, TransactionTestCase):
    def create_toggle(self):
        return waffle.get_waffle_flag_model().objects.create(
            name='transaction-flag-name', everyone=False
        )

    def flip_toggle(self, flag):
        flag.everyone = True
        flag.save()

    def toggle_is_active(self, flag):
        return waffle.flag_is_active(get(), flag.name)


class SwitchTransactionTests(TransactionTestMixin, TransactionTestCase):
    def create_toggle(self):
        return waffle.get_waffle_switch_model().objects.create(
            name="transaction-switch-name", active=False
        )

    def flip_toggle(self, switch):
        switch.active = True
        switch.save()

    def toggle_is_active(self, switch):
        return waffle.switch_is_active(switch.name)


class SampleTransactionTests(TransactionTestMixin, TransactionTestCase):
    def create_toggle(self):
        return waffle.get_waffle_sample_model().objects.create(
            name="transaction-sample-name", percent=0
        )

    def flip_toggle(self, sample):
        sample.percent = 100
        sample.save()

    def toggle_is_active(self, sample):
        return waffle.sample_is_active(sample.name)


class FunctionTests(TestCase):
    @mock.patch.object(random, 'uniform')
    def test_percent(self, uniform):
        """If you have no cookie, you get a cookie!"""
        uniform.return_value = '10'
        waffle.get_waffle_flag_model().objects.create(name='myflag', percent='50.0')
        request = get()
        response = process_request(request, views.flag_in_view)
        assert 'dwf_myflag' in response.cookies
        self.assertEqual('True', response.cookies['dwf_myflag'].value)
        self.assertEqual(b'on', response.content)

    @mock.patch.object(random, 'uniform')
    def test_percent_readonly(self, uniform):
        uniform.return_value = '10'
        """If you have no cookie, you do not get a cookie if we just have a question!"""
        waffle.get_waffle_flag_model().objects.create(name='myflag', percent='50.0')
        request = get()
        response = process_request(request, views.flag_in_view_readonly)
        assert 'dwf_myflag' not in response.cookies
        self.assertEqual(b'off', response.content)


class WaffleFlagEveryoneSettingTests(TestCase):
    databases = DATABASES

    def test_is_active_for_user_respects_everyone_on(self):
        """
        Test flag.is_active_for_user returns truthy value when everyone is set to True overriding all other settings.
        """
        flag = waffle.get_waffle_flag_model().objects.create(
            name="feature_flag_name",
            staff=False,
            everyone=True,
        )
        staff_user = get_user_model()(
            id=999,
            username="foo",
            is_staff=True,
        )
        self.assertTrue(flag.is_active_for_user(staff_user))

    def test_is_active_for_user_respects_everyone_off(self):
        """
        Test flag.is_active_for_user returns falsy value when everyone is set to False overriding all other settings.
        """
        flag = waffle.get_waffle_flag_model().objects.create(
            name="feature_flag_name",
            staff=True,
            everyone=False,
        )
        staff_user = get_user_model()(
            id=999,
            username="foo",
            is_staff=True,
        )
        self.assertFalse(flag.is_active_for_user(staff_user))



from django.urls import include, path
from django.contrib import admin
from django.http import HttpResponseNotFound, HttpResponseServerError

from test_app import views


def handler404(r, exception=None):
    return HttpResponseNotFound()


def handler500(r, exception=None):
    return HttpResponseServerError()


admin.autodiscover()

urlpatterns = [
    path('flag_in_view', views.flag_in_view, name='flag_in_view'),
    path('flag_in_view_readonly', views.flag_in_view_readonly, name='flag_in_view_readonly'),
    path('switch-on', views.switched_view),
    path('switch-off', views.switched_off_view),
    path('flag-on', views.flagged_view),
    path('foo_view', views.foo_view, name='foo_view'),
    path('foo_view_with_args/<int:some_number>/', views.foo_view_with_args, name='foo_view_with_args'),
    path('switched_view_with_valid_redirect',
         views.switched_view_with_valid_redirect),
    path('switched_view_with_valid_url_name',
         views.switched_view_with_valid_url_name),
    path('switched_view_with_args_with_valid_redirect/<int:some_number>/',
         views.switched_view_with_args_with_valid_redirect),
    path('switched_view_with_args_with_valid_url_name/<int:some_number>/',
         views.switched_view_with_args_with_valid_url_name),
    path('switched_view_with_invalid_redirect',
         views.switched_view_with_invalid_redirect),
    path('flagged_view_with_valid_redirect',
         views.flagged_view_with_valid_redirect),
    path('flagged_view_with_valid_url_name',
         views.flagged_view_with_valid_url_name),
    path('flagged_view_with_args_with_valid_redirect/<int:some_number>/',
         views.flagged_view_with_args_with_valid_redirect),
    path('flagged_view_with_args_with_valid_url_name/<int:some_number>/',
         views.flagged_view_with_args_with_valid_url_name),
    path('flagged_view_with_invalid_redirect',
         views.flagged_view_with_invalid_redirect),
    path('flag-off', views.flagged_off_view),
    path('', include('waffle.urls')),
    path('admin/', admin.site.urls),
]



from __future__ import annotations

import hashlib
from typing import Any

from django.conf import settings
from django.core.cache import BaseCache, caches

import waffle
from waffle import defaults


def get_setting(name: str, default: Any = None) -> Any:
    try:
        return getattr(settings, 'WAFFLE_' + name)
    except AttributeError:
        return getattr(defaults, name, default)


def keyfmt(k: str, v: str | None = None) -> str:
    prefix = get_setting('CACHE_PREFIX') + waffle.__version__
    if v is None:
        key = prefix + k
    else:
        key = prefix + hashlib.md5((k % v).encode('utf-8')).hexdigest()
    return key


def get_cache() -> BaseCache:
    CACHE_NAME = get_setting('CACHE_NAME')
    return caches[CACHE_NAME]



from __future__ import annotations

from functools import wraps, WRAPPER_ASSIGNMENTS
from typing import Any, Callable

from django.http import Http404, HttpRequest, HttpResponse, HttpResponsePermanentRedirect, HttpResponseRedirect
from django.shortcuts import redirect
from django.urls import reverse, NoReverseMatch

from waffle import flag_is_active, switch_is_active


def waffle_flag(
    flag_name: str, redirect_to: Callable | str | None = None,
) -> Callable[[Callable[[HttpRequest], HttpResponse]], Callable[[HttpRequest], HttpResponse]]:
    def decorator(view: Callable[[HttpRequest], HttpResponse]) -> Callable[[HttpRequest], HttpResponse]:
        @wraps(view, assigned=WRAPPER_ASSIGNMENTS)
        def _wrapped_view(request, *args, **kwargs):
            if flag_name.startswith('!'):
                active = not flag_is_active(request, flag_name[1:])
            else:
                active = flag_is_active(request, flag_name)

            if not active:
                response_to_redirect_to = get_response_to_redirect(redirect_to, *args, **kwargs)
                if response_to_redirect_to:
                    return response_to_redirect_to
                else:
                    raise Http404

            return view(request, *args, **kwargs)
        return _wrapped_view
    return decorator


def waffle_switch(
    switch_name: str, redirect_to: Callable | str | None = None,
) -> Callable[[Callable[[HttpRequest], HttpResponse]], Callable[[HttpRequest], HttpResponse]]:
    def decorator(view: Callable[[HttpRequest], HttpResponse]) -> Callable[[HttpRequest], HttpResponse]:
        @wraps(view, assigned=WRAPPER_ASSIGNMENTS)
        def _wrapped_view(request, *args, **kwargs):
            if switch_name.startswith('!'):
                active = not switch_is_active(switch_name[1:])
            else:
                active = switch_is_active(switch_name)

            if not active:
                response_to_redirect_to = get_response_to_redirect(redirect_to, *args, **kwargs)
                if response_to_redirect_to:
                    return response_to_redirect_to
                else:
                    raise Http404

            return view(request, *args, **kwargs)
        return _wrapped_view
    return decorator


def get_response_to_redirect(
    view: Callable | str | None, *args: Any, **kwargs: Any,
) -> HttpResponseRedirect | HttpResponsePermanentRedirect | None:
    try:
        return redirect(reverse(view, args=args, kwargs=kwargs)) if view else None
    except NoReverseMatch:
        return None



from typing import Generic, Optional, TypeVar, Union

from django.test.utils import TestContextDecorator

from waffle import (
    get_waffle_flag_model,
    get_waffle_sample_model,
    get_waffle_switch_model,
)


__all__ = ['override_flag', 'override_sample', 'override_switch']

_T = TypeVar("_T")


class _overrider(TestContextDecorator, Generic[_T]):
    def __init__(self, name: str, active: _T):
        super().__init__()
        self.name = name
        self.active = active

    def get(self) -> None:
        self.obj, self.created = self.cls.objects.get_or_create(name=self.name)

    def update(self, active: _T) -> None:
        raise NotImplementedError

    def get_value(self) -> _T:
        raise NotImplementedError

    def enable(self) -> None:
        self.get()
        self.old_value = self.get_value()
        if self.old_value != self.active:
            self.update(self.active)

    def disable(self) -> None:
        if self.created:
            self.obj.delete()
            self.obj.flush()
        else:
            self.update(self.old_value)


class override_switch(_overrider[bool]):
    """
    override_switch is a contextmanager for easier testing of switches.

    It accepts two parameters, name of the switch and it's state. Example
    usage::

        with override_switch('happy_mode', active=True):
            ...

    If `Switch` already existed, it's value would be changed inside the context
    block, then restored to the original value. If `Switch` did not exist
    before entering the context, it is created, then removed at the end of the
    block.

    It can also act as a decorator::

        @override_switch('happy_mode', active=True)
        def test_happy_mode_enabled():
            ...

    """
    cls = get_waffle_switch_model()

    def update(self, active: bool) -> None:
        obj = self.cls.objects.get(pk=self.obj.pk)
        obj.active = active
        obj.save()
        obj.flush()

    def get_value(self) -> bool:
        return self.obj.active


class override_flag(_overrider[Optional[bool]]):
    cls = get_waffle_flag_model()

    def update(self, active: Optional[bool]) -> None:
        obj = self.cls.objects.get(pk=self.obj.pk)
        obj.everyone = active
        obj.save()
        obj.flush()

    def get_value(self) -> Optional[bool]:
        return self.obj.everyone


class override_sample(_overrider[Union[bool, float]]):
    cls = get_waffle_sample_model()

    def get(self) -> None:
        try:
            self.obj = self.cls.objects.get(name=self.name)
            self.created = False
        except self.cls.DoesNotExist:
            self.obj = self.cls.objects.create(name=self.name, percent='0.0')
            self.created = True

    def update(self, active: Union[bool, float]) -> None:
        if active is True:
            p = 100.0
        elif active is False:
            p = 0.0
        else:
            p = active
        obj = self.cls.objects.get(pk=self.obj.pk)
        obj.percent = f'{p}'
        obj.save()
        obj.flush()

    def get_value(self) -> Union[bool, float]:
        p = self.obj.percent
        if p == 100.0:
            return True
        if p == 0.0:
            return False
        return p



from django.db.models.signals import m2m_changed
from django.dispatch import receiver

from waffle import get_waffle_flag_model


@receiver(m2m_changed, sender=get_waffle_flag_model().users.through)
@receiver(m2m_changed, sender=get_waffle_flag_model().groups.through)
def flag_membership_changed(sender, instance, action, **kwargs):
    if action in ('post_add', 'post_remove'):
        flag_model = get_waffle_flag_model()

        # instance could be a flag or an instance of the related model
        # https://docs.djangoproject.com/en/dev/ref/signals/#m2m-changed
        if isinstance(instance, flag_model):
            instance.flush()
        else:
            for flag in flag_model.objects.filter(pk__in=kwargs['pk_set']):
                flag.flush()



from typing import Any

from django.core.management.base import BaseCommand, CommandParser

from waffle import (
    get_waffle_flag_model,
    get_waffle_switch_model,
    get_waffle_sample_model,
)


class Command(BaseCommand):
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            '--flags',
            action='store',
            dest='flag_names',
            nargs='*',
            help='List of flag names to delete.')
        parser.add_argument(
            '--samples',
            action='store',
            dest='sample_names',
            nargs='*',
            help='List of sample names to delete.')
        parser.add_argument(
            '--switches',
            action='store',
            dest='switch_names',
            nargs='*',
            help='List of switch names to delete.')

    help = 'Delete flags, samples, and switches from database'

    def handle(self, *args: Any, **options: Any) -> None:
        flags = options['flag_names']
        if flags:
            flag_queryset = get_waffle_flag_model().objects.filter(name__in=flags)
            flag_count = flag_queryset.count()
            flag_queryset.delete()
            self.stdout.write(f'Deleted {flag_count} Flags')

        switches = options['switch_names']
        if switches:
            switches_queryset = get_waffle_switch_model().objects.filter(
                name__in=switches
            )
            switch_count = switches_queryset.count()
            switches_queryset.delete()
            self.stdout.write(f'Deleted {switch_count} Switches')

        samples = options['sample_names']
        if samples:
            sample_queryset = get_waffle_sample_model().objects.filter(name__in=samples)
            sample_count = sample_queryset.count()
            sample_queryset.delete()
            self.stdout.write(f'Deleted {sample_count} Samples')



from django.urls import reverse

from waffle import (
    get_waffle_flag_model,
    get_waffle_sample_model,
    get_waffle_switch_model,
)
from waffle.tests.base import TestCase


class WaffleViewTests(TestCase):
    def test_wafflejs(self):
        response = self.client.get(reverse('wafflejs'))
        self.assertEqual(200, response.status_code)
        self.assertEqual('application/x-javascript', response['content-type'])
        cache_control = [control.strip()
                         for control in response['cache-control'].split(',')]
        self.assertIn('max-age=0', cache_control)

    def test_waffle_status(self):
        response = self.client.get(reverse('waffle_status'))
        self.assertEqual(200, response.status_code)
        self.assertEqual('application/json', response['content-type'])
        cache_control = [control.strip()
                         for control in response['cache-control'].split(',')]
        self.assertIn('max-age=0', cache_control)

    def test_waffle_status_response(self):
        get_waffle_flag_model().objects.create(name='test_flag_active', everyone=True)
        get_waffle_flag_model().objects.create(name='test_flag_inactive', everyone=False)
        get_waffle_switch_model().objects.create(name='test_switch_active', active=True)
        get_waffle_switch_model().objects.create(name='test_switch_inactive', active=False)
        get_waffle_sample_model().objects.create(name='test_sample_active', percent=100)
        get_waffle_sample_model().objects.create(name='test_sample_inactive', percent=0)

        response = self.client.get(reverse('waffle_status'))
        self.assertEqual(200, response.status_code)
        content = response.json()

        assert 'test_flag_active' in content['flags'].keys()
        assert content['flags']['test_flag_active']['is_active']
        assert 'test_flag_inactive' in content['flags'].keys()
        assert not content['flags']['test_flag_inactive']['is_active']

        assert 'test_switch_active' in content['switches'].keys()
        assert content['switches']['test_switch_active']['is_active']
        assert 'test_switch_inactive' in content['switches'].keys()
        assert not content['switches']['test_switch_inactive']['is_active']

        assert 'test_sample_active' in content['samples'].keys()
        assert content['samples']['test_sample_active']['is_active']
        assert 'test_sample_inactive' in content['samples'].keys()
        assert not content['samples']['test_sample_inactive']['is_active']

    def test_flush_all_flags(self):
        """Test the 'FLAGS_ALL' list gets invalidated correctly."""
        get_waffle_flag_model().objects.create(name='myflag1', everyone=True)
        response = self.client.get(reverse('wafflejs'))
        self.assertEqual(200, response.status_code)
        assert ('myflag1', True) in response.context['flags']

        get_waffle_flag_model().objects.create(name='myflag2', everyone=True)
        response = self.client.get(reverse('wafflejs'))
        self.assertEqual(200, response.status_code)
        assert ('myflag2', True) in response.context['flags']

    def test_flush_all_switches(self):
        """Test the 'SWITCHES_ALL' list gets invalidated correctly."""
        switch = get_waffle_switch_model().objects.create(name='myswitch', active=True)
        response = self.client.get(reverse('wafflejs'))
        self.assertEqual(200, response.status_code)
        assert ('myswitch', True) in response.context['switches']

        switch.active = False
        switch.save()
        response = self.client.get(reverse('wafflejs'))
        self.assertEqual(200, response.status_code)
        assert ('myswitch', False) in response.context['switches']

    def test_flush_all_samples(self):
        """Test the 'SAMPLES_ALL' list gets invalidated correctly."""
        get_waffle_sample_model().objects.create(name='sample1', percent='100.0')
        response = self.client.get(reverse('wafflejs'))
        self.assertEqual(200, response.status_code)
        assert ('sample1', True) in response.context['samples']

        get_waffle_sample_model().objects.create(name='sample2', percent='100.0')

        response = self.client.get(reverse('wafflejs'))
        self.assertEqual(200, response.status_code)
        assert ('sample2', True) in response.context['samples']



# Generated by Django 3.0.9 on 2020-08-18 04:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('waffle', '0003_update_strings_for_i18n'),
    ]

    operations = [
        migrations.AlterField(
            model_name='flag',
            name='everyone',
            field=models.BooleanField(blank=True, help_text='Flip this flag on (Yes) or off (No) for everyone, overriding all other settings. Leave as Unknown to use normally.', null=True, verbose_name='Everyone'),
        ),
    ]



from django.http import HttpResponse
from django.shortcuts import render
from django.template.loader import render_to_string
from django.views.generic import View

from waffle import flag_is_active
from waffle.decorators import waffle_flag, waffle_switch
from waffle.mixins import WaffleFlagMixin, WaffleSampleMixin, WaffleSwitchMixin


def flag_in_view(request):
    if flag_is_active(request, 'myflag'):
        return HttpResponse('on')
    return HttpResponse('off')


def flag_in_view_readonly(request):
    if flag_is_active(request, 'myflag', read_only=True):
        return HttpResponse('on')
    return HttpResponse('off')


def flag_in_jinja(request):
    return render(request, 'jinja/jinja.html')


def flag_in_django(request):
    context = {
        'flag_var': 'flag_var',
        'switch_var': 'switch_var',
        'sample_var': 'sample_var',
    }
    return render(request, 'django/django.html', context=context)


def no_request_context(request):
    return render_to_string('django/django_email.html', context={})


@waffle_switch('foo')
def switched_view(request):
    return HttpResponse('foo')


@waffle_switch('!foo')
def switched_off_view(request):
    return HttpResponse('foo')


@waffle_flag('foo')
def flagged_view(request):
    return HttpResponse('foo')


@waffle_flag('!foo')
def flagged_off_view(request):
    return HttpResponse('foo')


def foo_view(request):
    return HttpResponse('redirected')


def foo_view_with_args(request, some_number):
    return HttpResponse(f'redirected with {some_number}')


@waffle_switch('foo', redirect_to=foo_view)
def switched_view_with_valid_redirect(request):
    return HttpResponse('foo')


@waffle_switch('foo', redirect_to='foo_view')
def switched_view_with_valid_url_name(request):
    return HttpResponse('foo')


@waffle_switch('foo', redirect_to=foo_view_with_args)
def switched_view_with_args_with_valid_redirect(request, some_number):
    return HttpResponse(f'foo with {some_number}')


@waffle_switch('foo', redirect_to='foo_view_with_args')
def switched_view_with_args_with_valid_url_name(request, some_number):
    return HttpResponse(f'foo with {some_number}')


@waffle_switch('foo', redirect_to='invalid_view')
def switched_view_with_invalid_redirect(request):
    return HttpResponse('foo')


@waffle_flag('foo', redirect_to=foo_view)
def flagged_view_with_valid_redirect(request):
    return HttpResponse('foo')


@waffle_flag('foo', redirect_to='foo_view')
def flagged_view_with_valid_url_name(request):
    return HttpResponse('foo')


@waffle_flag('foo', redirect_to=foo_view_with_args)
def flagged_view_with_args_with_valid_redirect(request, some_number):
    return HttpResponse(f'foo with {some_number}')


@waffle_flag('foo', redirect_to='foo_view_with_args')
def flagged_view_with_args_with_valid_url_name(request, some_number):
    return HttpResponse(f'foo with {some_number}')


@waffle_flag('foo', redirect_to='invalid_view')
def flagged_view_with_invalid_redirect(request):
    return HttpResponse('foo')


class BaseWaffleView(View):
    def get(self, request, *args, **kwargs):
        return HttpResponse('foo')


class FlagView(WaffleFlagMixin, BaseWaffleView):
    waffle_flag = 'foo'


class FlagOffView(WaffleFlagMixin, BaseWaffleView):
    waffle_flag = '!foo'


class SampleView(WaffleSampleMixin, BaseWaffleView):
    waffle_sample = 'foo'


class SampleOffView(WaffleSampleMixin, BaseWaffleView):
    waffle_sample = '!foo'


class SwitchView(WaffleSwitchMixin, BaseWaffleView):
    waffle_switch = 'foo'


class SwitchOffView(WaffleSwitchMixin, BaseWaffleView):
    waffle_switch = '!foo'



from __future__ import annotations

from typing import Any

from django.http import HttpRequest, HttpResponse, JsonResponse
from django.template import loader
from django.views.decorators.cache import never_cache

from waffle import get_waffle_flag_model, get_waffle_switch_model, get_waffle_sample_model
from waffle.utils import get_setting


@never_cache
def wafflejs(request):
    return HttpResponse(_generate_waffle_js(request),
                        content_type='application/x-javascript')


def _generate_waffle_js(request: HttpRequest) -> str:
    flags = get_waffle_flag_model().get_all()
    flag_values = [(f.name, f.is_active(request)) for f in flags]

    switches = get_waffle_switch_model().get_all()
    switch_values = [(s.name, s.is_active()) for s in switches]

    samples = get_waffle_sample_model().get_all()
    sample_values = [(s.name, s.is_active()) for s in samples]

    return loader.render_to_string('waffle/waffle.js', {
        'flags': flag_values,
        'switches': switch_values,
        'samples': sample_values,
        'flag_default': get_setting('FLAG_DEFAULT'),
        'switch_default': get_setting('SWITCH_DEFAULT'),
        'sample_default': get_setting('SAMPLE_DEFAULT'),
    })


@never_cache
def waffle_json(request):
    return JsonResponse(_generate_waffle_json(request))


def _generate_waffle_json(request: HttpRequest) -> dict[str, dict[str, Any]]:
    flags = get_waffle_flag_model().get_all()
    flag_values = {
        f.name: {
            'is_active': f.is_active(request),
            'last_modified': f.modified,
        }
        for f in flags
    }

    switches = get_waffle_switch_model().get_all()
    switch_values = {
        s.name: {
            'is_active': s.is_active(),
            'last_modified': s.modified,
        }
        for s in switches
    }

    samples = get_waffle_sample_model().get_all()
    sample_values = {
        s.name: {
            'is_active': s.is_active(),
            'last_modified': s.modified,
        }
        for s in samples
    }

    return {
        'flags': flag_values,
        'switches': switch_values,
        'samples': sample_values,
    }



from typing import Any

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.core.management.base import BaseCommand, CommandError, CommandParser
from django.db.models import Q

from waffle.models import AbstractBaseFlag
from waffle import get_waffle_flag_model

UserModel = get_user_model()


class Command(BaseCommand):
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            'name',
            nargs='?',
            help='The name of the flag.')
        parser.add_argument(
            '-l', '--list',
            action='store_true',
            dest='list_flags',
            default=False,
            help='List existing samples.')
        parser.add_argument(
            '--everyone',
            action='store_true',
            dest='everyone',
            help='Activate flag for all users.')
        parser.add_argument(
            '--deactivate',
            action='store_false',
            dest='everyone',
            help='Deactivate flag for all users.')
        parser.add_argument(
            '--percent', '-p',
            action='store',
            type=int,
            dest='percent',
            help='Roll out the flag for a certain percentage of users. Takes '
                 'a number between 0.0 and 100.0')
        parser.add_argument(
            '--superusers',
            action='store_true',
            dest='superusers',
            default=False,
            help='Turn on the flag for Django superusers.')
        parser.add_argument(
            '--staff',
            action='store_true',
            dest='staff',
            default=False,
            help='Turn on the flag for Django staff.')
        parser.add_argument(
            '--authenticated',
            action='store_true',
            dest='authenticated',
            default=False,
            help='Turn on the flag for logged in users.')
        parser.add_argument(
            '--group', '-g',
            action='append',
            default=list(),
            help='Turn on the flag for listed group names (use flag more '
                 'than once for multiple groups). WARNING: This will remove '
                 'any currently associated groups unless --append is used!')
        parser.add_argument(
            '--user', '-u',
            action='append',
            default=list(),
            help='Turn on the flag for listed usernames (use flag more '
                 'than once for multiple users). WARNING: This will remove '
                 'any currently associated users unless --append is used!')
        parser.add_argument(
            '--append',
            action='store_true',
            dest='append',
            default=False,
            help='Append only mode when adding groups.')
        parser.add_argument(
            '--rollout', '-r',
            action='store_true',
            dest='rollout',
            default=False,
            help='Turn on rollout mode.')
        parser.add_argument(
            '--testing', '-t',
            action='store_true',
            dest='testing',
            default=False,
            help='Turn on testing mode, allowing the flag to be specified via '
                 'a querystring parameter.')
        parser.add_argument(
            '--create',
            action='store_true',
            dest='create',
            default=False,
            help='If the flag doesn\'t exist, create it.')
        parser.set_defaults(everyone=None)

    help = 'Modify a flag.'

    def handle(self, *args: Any, **options: Any) -> None:
        if options['list_flags']:
            self.stdout.write('Flags:')
            for flag in get_waffle_flag_model().objects.iterator():
                self.log_flag_to_stdout(flag)
            return

        flag_name = options['name']

        if not flag_name:
            raise CommandError('You need to specify a flag name.')

        if options['create']:
            flag, created = get_waffle_flag_model().objects.get_or_create(name=flag_name)
            if created:
                self.stdout.write(f'Creating flag: {flag_name}')
        else:
            try:
                flag = get_waffle_flag_model().objects.get(name=flag_name)
            except get_waffle_flag_model().DoesNotExist:
                raise CommandError('This flag does not exist.')

        # Group isn't an attribute on the Flag, but a related Many-to-Many
        # field, so we handle it a bit differently by looking up groups and
        # adding each group to the flag individually
        options_append = options.pop('append')
        if groups := options.pop('group'):
            group_hash = {}
            for group in groups:
                try:
                    group_instance = Group.objects.get(name=group)
                    group_hash[group_instance.name] = group_instance.id
                except Group.DoesNotExist:
                    raise CommandError(f'Group {group} does not exist')
            # If 'append' was not passed, we clear related groups
            if not options_append:
                flag.groups.clear()
            self.stdout.write('Setting group(s): %s' % (
                [name for name, _id in group_hash.items()])
                              )
            for group_id in group_hash.values():
                flag.groups.add(group_id)
        if users := options.pop('user'):
            user_hash = set()
            for username in users:
                try:
                    user_instance = UserModel.objects.get(
                        Q(**{UserModel.USERNAME_FIELD: username})
                        | Q(**{UserModel.EMAIL_FIELD: username})
                    )
                    user_hash.add(user_instance)
                except UserModel.DoesNotExist:
                    raise CommandError(f'User {username} does not exist')
            # If 'append' was not passed, we clear related users
            if not options_append:
                flag.users.clear()
            self.stdout.write(f'Setting user(s): {user_hash}')
            # for user in user_hash:
            flag.users.add(*[user.id for user in user_hash])
        for option_name, option in options.items():
            if hasattr(flag, option_name):
                self.stdout.write(f'Setting {option_name}: {option}')
                setattr(flag, option_name, option)

        flag.save()

    def log_flag_to_stdout(self, flag: AbstractBaseFlag) -> None:
        self.stdout.write(f'NAME: {flag.name}')
        self.stdout.write(f'SUPERUSERS: {flag.superusers}')
        self.stdout.write(f'EVERYONE: {flag.everyone}')
        self.stdout.write(f'AUTHENTICATED: {flag.authenticated}')
        self.stdout.write(f'PERCENT: {flag.percent}')
        self.stdout.write(f'TESTING: {flag.testing}')
        self.stdout.write(f'ROLLOUT: {flag.rollout}')
        self.stdout.write(f'STAFF: {flag.staff}')
        self.stdout.write('GROUPS: {}'.format(list(
            flag.groups.values_list('name', flat=True)))
        )
        self.stdout.write('USERS: {}'.format(list(
            flag.users.values_list(UserModel.USERNAME_FIELD, flat=True)))
        )
        self.stdout.write('')



# Generated by Django 2.0.4 on 2018-04-26 07:42

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Company',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name='CompanyAwareFlag',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='The human/computer readable name.', max_length=100, unique=True, verbose_name='Name')),
                ('everyone', models.NullBooleanField(help_text='Flip this flag on (Yes) or off (No) for everyone, overriding all other settings. Leave as Unknown to use normally.', verbose_name='Everyone')),
                ('percent', models.DecimalField(blank=True, decimal_places=1, help_text='A number between 0.0 and 99.9 to indicate a percentage of users for whom this flag will be active.', max_digits=3, null=True, verbose_name='Percent')),
                ('testing', models.BooleanField(default=False, help_text='Allow this flag to be set for a session for user testing', verbose_name='Testing')),
                ('superusers', models.BooleanField(default=True, help_text='Flag always active for superusers?', verbose_name='Superusers')),
                ('staff', models.BooleanField(default=False, help_text='Flag always active for staff?', verbose_name='Staff')),
                ('authenticated', models.BooleanField(default=False, help_text='Flag always active for authenticated users?', verbose_name='Authenticated')),
                ('languages', models.TextField(blank=True, default='', help_text='Activate this flag for users with one of these languages (comma-separated list)', verbose_name='Languages')),
                ('rollout', models.BooleanField(default=False, help_text='Activate roll-out mode?', verbose_name='Rollout')),
                ('note', models.TextField(blank=True, help_text='Note where this Flag is used.', verbose_name='Note')),
                ('created', models.DateTimeField(db_index=True, default=django.utils.timezone.now, help_text='Date when this Flag was created.', verbose_name='Created')),
                ('modified', models.DateTimeField(default=django.utils.timezone.now, help_text='Date when this Flag was last modified.', verbose_name='Modified')),
                ('companies', models.ManyToManyField(blank=True, help_text='Activate this flag for these companies.', to='test_app.Company')),
                ('groups', models.ManyToManyField(blank=True, help_text='Activate this flag for these user groups.', to='auth.Group', verbose_name='Groups')),
                ('users', models.ManyToManyField(blank=True, help_text='Activate this flag for these users.', to=settings.AUTH_USER_MODEL, verbose_name='Users')),
            ],
            options={
                'verbose_name': 'Flag',
                'verbose_name_plural': 'Flags',
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='CompanyUser',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('username', models.CharField(max_length=100)),
                ('company', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='test_app.Company')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from django.apps import AppConfig


class WaffleConfig(AppConfig):
    name = 'waffle'
    verbose_name = 'django-waffle'
    default_auto_field = 'django.db.models.AutoField'

    def ready(self) -> None:
        import waffle.signals  # noqa: F401



from django.urls import path

from waffle.views import wafflejs, waffle_json

urlpatterns = [
    path('wafflejs', wafflejs, name='wafflejs'),
    path('waffle_status', waffle_json, name='waffle_status'),
]



import os


try:
    import django_jinja
    JINJA_INSTALLED = True
except ImportError:
    JINJA_INSTALLED = False


# Make filepaths relative to settings.
ROOT = os.path.dirname(os.path.abspath(__file__))
path = lambda *a: os.path.join(ROOT, *a)  # noqa: E731

DEBUG = True
TEST_RUNNER = 'django.test.runner.DiscoverRunner'

JINJA_CONFIG = {}

SITE_ID = 1
USE_I18N = False

SECRET_KEY = 'foobar'

DATABASES = {
    'default': {
        'NAME': 'test.db',
        'ENGINE': 'django.db.backends.sqlite3',
    },

    # Provide a readonly DB for testing DB replication scenarios.
    'readonly': {
        'NAME': 'test.readonly.db',
        'ENGINE': 'django.db.backends.sqlite3',
    }
}

DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'waffle',
    'test_app',
)

MIDDLEWARE = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'waffle.middleware.WaffleMiddleware',
)


ROOT_URLCONF = 'test_app.urls'

_CONTEXT_PROCESSORS = (
    'django.contrib.auth.context_processors.auth',
    'django.template.context_processors.request',
    'django.contrib.messages.context_processors.messages',
)


if JINJA_INSTALLED:
    TEMPLATES = [
        {
            'BACKEND': 'django_jinja.backend.Jinja2',
            'DIRS': [],
            'APP_DIRS': True,
            'OPTIONS': {
                'match_regex': r'jinja.*',
                'match_extension': '',
                'newstyle_gettext': True,
                'context_processors': _CONTEXT_PROCESSORS,
                'undefined': 'jinja2.Undefined',
                'extensions': [
                    'jinja2.ext.i18n',
                    'waffle.jinja.WaffleExtension',
                ],
            }
        }
    ]
else:
    TEMPLATES = []

TEMPLATES.append(
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'debug': DEBUG,
            'context_processors': _CONTEXT_PROCESSORS,
        }
    }
)


WAFFLE_FLAG_DEFAULT = False
WAFFLE_SWITCH_DEFAULT = False
WAFFLE_SAMPLE_DEFAULT = False
WAFFLE_READ_FROM_WRITE_DB = False
WAFFLE_OVERRIDE = False
WAFFLE_ENABLE_ADMIN_PAGES = True
WAFFLE_CACHE_PREFIX = 'test:'



from decimal import Decimal

from django.contrib.auth.models import AnonymousUser
from django.db import transaction
from django.test import TransactionTestCase, RequestFactory, TestCase

import waffle
from waffle.testutils import override_switch, override_flag, override_sample


class OverrideSwitchMixin:
    def test_switch_existed_and_was_active(self):
        waffle.get_waffle_switch_model().objects.create(name='foo', active=True)

        with override_switch('foo', active=True):
            assert waffle.switch_is_active('foo')

        with override_switch('foo', active=False):
            assert not waffle.switch_is_active('foo')

        # make sure it didn't change 'active' value
        assert waffle.get_waffle_switch_model().objects.get(name='foo').active

    def test_switch_existed_and_was_NOT_active(self):
        waffle.get_waffle_switch_model().objects.create(name='foo', active=False)

        with override_switch('foo', active=True):
            assert waffle.switch_is_active('foo')

        with override_switch('foo', active=False):
            assert not waffle.switch_is_active('foo')

        # make sure it didn't change 'active' value
        assert not waffle.get_waffle_switch_model().objects.get(name='foo').active

    def test_new_switch(self):
        assert not waffle.get_waffle_switch_model().objects.filter(name='foo').exists()

        with override_switch('foo', active=True):
            assert waffle.switch_is_active('foo')

        with override_switch('foo', active=False):
            assert not waffle.switch_is_active('foo')

        assert not waffle.get_waffle_switch_model().objects.filter(name='foo').exists()

    def test_as_decorator(self):
        assert not waffle.get_waffle_switch_model().objects.filter(name='foo').exists()

        @override_switch('foo', active=True)
        def test_enabled():
            assert waffle.switch_is_active('foo')

        test_enabled()

        @override_switch('foo', active=False)
        def test_disabled():
            assert not waffle.switch_is_active('foo')

        test_disabled()

        assert not waffle.get_waffle_switch_model().objects.filter(name='foo').exists()

    def test_restores_after_exception(self):
        waffle.get_waffle_switch_model().objects.create(name='foo', active=True)

        def inner():
            with override_switch('foo', active=False):
                raise RuntimeError("Trying to break")

        with self.assertRaises(RuntimeError):
            inner()

        assert waffle.get_waffle_switch_model().objects.get(name='foo').active

    def test_restores_after_exception_in_decorator(self):
        waffle.get_waffle_switch_model().objects.create(name='foo', active=True)

        @override_switch('foo', active=False)
        def inner():
            raise RuntimeError("Trying to break")

        with self.assertRaises(RuntimeError):
            inner()

        assert waffle.get_waffle_switch_model().objects.get(name='foo').active

    def test_cache_is_flushed_by_testutils_even_in_transaction(self):
        waffle.get_waffle_switch_model().objects.create(name='foo', active=True)

        with transaction.atomic():
            with override_switch('foo', active=True):
                assert waffle.switch_is_active('foo')

            with override_switch('foo', active=False):
                assert not waffle.switch_is_active('foo')

        assert waffle.switch_is_active('foo')


class OverrideSwitchTestCase(OverrideSwitchMixin, TestCase):
    """
    Run tests with Django TestCase
    """


class OverrideSwitchTransactionTestCase(OverrideSwitchMixin, TransactionTestCase):
    """
    Run tests with Django TransactionTestCase
    """


def req():
    r = RequestFactory().get('/')
    r.user = AnonymousUser()
    return r


class OverrideFlagTestsMixin:
    def test_flag_existed_and_was_active(self):
        waffle.get_waffle_flag_model().objects.create(name='foo', everyone=True)

        with override_flag('foo', active=True):
            assert waffle.flag_is_active(req(), 'foo')

        with override_flag('foo', active=False):
            assert not waffle.flag_is_active(req(), 'foo')

        assert waffle.get_waffle_flag_model().objects.get(name='foo').everyone

    def test_flag_existed_and_was_inactive(self):
        waffle.get_waffle_flag_model().objects.create(name='foo', everyone=False)

        with override_flag('foo', active=True):
            assert waffle.flag_is_active(req(), 'foo')

        with override_flag('foo', active=False):
            assert not waffle.flag_is_active(req(), 'foo')

        assert waffle.get_waffle_flag_model().objects.get(name='foo').everyone is False

    def test_flag_existed_and_was_null(self):
        waffle.get_waffle_flag_model().objects.create(name='foo', everyone=None)

        with override_flag('foo', active=True):
            assert waffle.flag_is_active(req(), 'foo')

        with override_flag('foo', active=False):
            assert not waffle.flag_is_active(req(), 'foo')

        assert waffle.get_waffle_flag_model().objects.get(name='foo').everyone is None

    def test_flag_did_not_exist(self):
        assert not waffle.get_waffle_flag_model().objects.filter(name='foo').exists()

        with override_flag('foo', active=True):
            assert waffle.flag_is_active(req(), 'foo')

        with override_flag('foo', active=False):
            assert not waffle.flag_is_active(req(), 'foo')

        assert not waffle.get_waffle_flag_model().objects.filter(name='foo').exists()

    def test_cache_is_flushed_by_testutils_even_in_transaction(self):
        waffle.get_waffle_flag_model().objects.create(name='foo', everyone=True)

        with transaction.atomic():
            with override_flag('foo', active=True):
                assert waffle.flag_is_active(req(), 'foo')

            with override_flag('foo', active=False):
                assert not waffle.flag_is_active(req(), 'foo')

        assert waffle.flag_is_active(req(), 'foo')


class OverrideFlagsTestCase(OverrideFlagTestsMixin, TestCase):
    """
    Run tests with Django TestCase
    """


class OverrideFlagsTransactionTestCase(OverrideFlagTestsMixin, TransactionTestCase):
    """
    Run tests with Django TransactionTestCase
    """


class OverrideSampleTestsMixin:
    def test_sample_existed_and_was_100(self):
        waffle.get_waffle_sample_model().objects.create(name='foo', percent='100.0')

        with override_sample('foo', active=True):
            assert waffle.sample_is_active('foo')

        with override_sample('foo', active=False):
            assert not waffle.sample_is_active('foo')

        self.assertEqual(Decimal('100.0'),
                         waffle.get_waffle_sample_model().objects.get(name='foo').percent)

    def test_sample_existed_and_was_0(self):
        waffle.get_waffle_sample_model().objects.create(name='foo', percent='0.0')

        with override_sample('foo', active=True):
            assert waffle.sample_is_active('foo')

        with override_sample('foo', active=False):
            assert not waffle.sample_is_active('foo')

        self.assertEqual(Decimal('0.0'),
                         waffle.get_waffle_sample_model().objects.get(name='foo').percent)

    def test_sample_existed_and_was_50(self):
        waffle.get_waffle_sample_model().objects.create(name='foo', percent='50.0')

        with override_sample('foo', active=True):
            assert waffle.sample_is_active('foo')

        with override_sample('foo', active=False):
            assert not waffle.sample_is_active('foo')

        self.assertEqual(Decimal('50.0'),
                         waffle.get_waffle_sample_model().objects.get(name='foo').percent)

    def test_sample_did_not_exist(self):
        assert not waffle.get_waffle_sample_model().objects.filter(name='foo').exists()

        with override_sample('foo', active=True):
            assert waffle.sample_is_active('foo')

        with override_sample('foo', active=False):
            assert not waffle.sample_is_active('foo')

        assert not waffle.get_waffle_sample_model().objects.filter(name='foo').exists()

    def test_cache_is_flushed_by_testutils_even_in_transaction(self):
        waffle.get_waffle_sample_model().objects.create(name='foo', percent='100.0')

        with transaction.atomic():
            with override_sample('foo', active=True):
                assert waffle.sample_is_active('foo')

            with override_sample('foo', active=False):
                assert not waffle.sample_is_active('foo')

        assert waffle.sample_is_active('foo')


class OverrideSampleTestCase(OverrideSampleTestsMixin, TestCase):
    """
    Run tests with Django TestCase
    """


class OverrideSampleTransactionTestCase(OverrideSampleTestsMixin, TransactionTestCase):
    """
    Run tests with Django TransactionTestCase
    """


class OverrideSwitchOnClassTestsMixin:
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        assert not waffle.get_waffle_switch_model().objects.filter(name='foo').exists()
        waffle.get_waffle_switch_model().objects.create(name='foo', active=True)

    def test_undecorated_method_is_set_properly_for_switch(self):
        self.assertFalse(waffle.switch_is_active('foo'))


@override_switch('foo', active=False)
class OverrideSwitchOnClassTestCase(OverrideSwitchOnClassTestsMixin,
                                    TestCase):
    """
    Run tests with Django TestCase
    """


@override_switch('foo', active=False)
class OverrideSwitchOnClassTransactionTestCase(OverrideSwitchOnClassTestsMixin,
                                               TransactionTestCase):
    """
    Run tests with Django TransactionTestCase
    """


class OverrideFlagOnClassTestsMixin:
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        assert not waffle.get_waffle_flag_model().objects.filter(name='foo').exists()
        waffle.get_waffle_flag_model().objects.create(name='foo', everyone=True)

    def test_undecorated_method_is_set_properly_for_flag(self):
        self.assertFalse(waffle.flag_is_active(req(), 'foo'))


@override_flag('foo', active=False)
class OverrideFlagOnClassTestCase(OverrideFlagOnClassTestsMixin,
                                  TestCase):
    """
    Run tests with Django TestCase
    """


@override_flag('foo', active=False)
class OverrideFlagOnClassTransactionTestCase(OverrideFlagOnClassTestsMixin,
                                             TransactionTestCase):
    """
    Run tests with Django TransactionTestCase
    """


class OverrideSampleOnClassTestsMixin:
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        assert not waffle.get_waffle_sample_model().objects.filter(name='foo').exists()
        waffle.get_waffle_sample_model().objects.create(name='foo', percent='100.0')

    def test_undecorated_method_is_set_properly_for_sample(self):
        self.assertFalse(waffle.sample_is_active('foo'))


@override_sample('foo', active=False)
class OverrideSampleOnClassTestCase(OverrideSampleOnClassTestsMixin,
                                    TestCase):
    """
    Run tests with Django TestCase
    """


@override_sample('foo', active=False)
class OverrideSampleOnClassTransactionTestCase(OverrideSampleOnClassTestsMixin,
                                               TransactionTestCase):
    """
    Run tests with Django TransactionTestCase
    """


class InheritanceOverrideSwitchOnClassTests(OverrideSwitchOnClassTestCase):
    """
    Extend ``OverrideSwitchOnClassTestCase``
    and make sure ``override_switch`` change still works.
    """

    def test_child_undecorated_method_is_set_properly_for_switch(self):
        self.assertFalse(waffle.switch_is_active('foo'))


class InheritanceOverrideFlagOnClassTests(OverrideFlagOnClassTestCase):
    """
    Extend ``OverrideFlagOnClassTestCase``
    and make sure ``override_flag`` change still works.
    """

    def test_child_undecorated_method_is_set_properly_for_flag(self):
        self.assertFalse(waffle.flag_is_active(req(), 'foo'))


class InheritanceOverrideSampleOnClassTests(OverrideSampleOnClassTestCase):
    """
    Extend ``OverrideSampleOnClassTestCase``
    and make sure ``override_sample`` change still works.
    """

    def test_child_undecorated_method_is_set_properly_for_sample(self):
        self.assertFalse(waffle.sample_is_active('foo'))



from django.contrib.auth.base_user import AbstractBaseUser
from django.db import models
from django.db.models import CASCADE
from django.utils.translation import gettext_lazy as _

from waffle.models import AbstractBaseSample, AbstractBaseSwitch, AbstractUserFlag, CACHE_EMPTY
from waffle.utils import get_setting, keyfmt, get_cache

cache = get_cache()


class Company(models.Model):
    name = models.CharField(
        max_length=100,
    )


class CompanyUser(AbstractBaseUser):
    company = models.ForeignKey(
        Company,
        on_delete=CASCADE
    )

    username = models.CharField(
        max_length=100,
    )


class CustomSwitch(AbstractBaseSwitch):
    """Demonstrates custom switch behavior."""


class CustomSample(AbstractBaseSample):
    """Demonstrates custom switch behavior."""


class CompanyAwareFlag(AbstractUserFlag):
    FLAG_COMPANIES_CACHE_KEY = 'FLAG_COMPANIES_CACHE_KEY'
    FLAG_COMPANIES_CACHE_KEY_DEFAULT = 'flag:%s:companies'

    companies = models.ManyToManyField(
        Company,
        blank=True,
        help_text=_('Activate this flag for these companies.'),
    )

    def get_flush_keys(self, flush_keys=None):
        flush_keys = super().get_flush_keys(flush_keys)
        companies_cache_key = get_setting(CompanyAwareFlag.FLAG_COMPANIES_CACHE_KEY,
                                          CompanyAwareFlag.FLAG_COMPANIES_CACHE_KEY_DEFAULT)
        flush_keys.append(keyfmt(companies_cache_key, self.name))
        return flush_keys

    def is_active_for_user(self, user):
        is_active = super().is_active_for_user(user)
        if is_active:
            return is_active

        if getattr(user, 'company_id', None):
            company_ids = self._get_company_ids()
            if user.company_id in company_ids:
                return True

    def _get_company_ids(self):
        cache_key = keyfmt(
            get_setting(CompanyAwareFlag.FLAG_COMPANIES_CACHE_KEY, CompanyAwareFlag.FLAG_COMPANIES_CACHE_KEY_DEFAULT),
            self.name
        )
        cached = cache.get(cache_key)
        if cached == CACHE_EMPTY:
            return set()
        if cached:
            return cached

        company_ids = set(self.companies.all().values_list('pk', flat=True))
        if not company_ids:
            cache.add(cache_key, CACHE_EMPTY)
            return set()

        cache.add(cache_key, company_ids)
        return company_ids
